import { Injectable, Directive, TemplateRef, ContentChild, Input, Component, ChangeDetectionStrategy, Output, EventEmitter, ElementRef, NgZone, ViewChild, ChangeDetectorRef, ViewEncapsulation, ContentChildren, ViewContainerRef, NgModule } from '@angular/core';
import { combineLatest, BehaviorSubject } from 'rxjs';
import { debounceTime } from 'rxjs/operators';
import { NgbModule } from '@ng-bootstrap/ng-bootstrap';
import { CommonModule } from '@angular/common';
import { FormsModule } from '@angular/forms';
import { ResizableModule } from 'angular-resizable-element';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
const startCase = require('lodash/startCase');
const toLower = require('lodash/toLower');
class DataGridService {
    constructor() { }
    /**
     * Maps external table/column/control properties to those needed by the datatable
     * @param {?} array
     * @param {?} mapObj
     * @return {?}
     */
    mapPropertiesDown(array, mapObj) {
        return array.map(element => {
            for (const /** @type {?} */ key in mapObj) {
                if (mapObj.hasOwnProperty(key)) {
                    // If the default property needed by the DT is NOT found in the object
                    // This avoids issues where the correct property is actually being passed in even though it is mapped
                    if (!element[key]) {
                        // If mapping data, any column references will be camel case instead of title case. Remap to camel case
                        element[key] = element[mapObj[key]];
                        // delete element[mapObj[key]];
                    }
                }
            }
            return element;
        });
    }
    /**
     * If external map properties are specified, map them back to the input types before emitting them up to the host component
     * @param {?} array
     * @param {?} mapObj
     * @return {?}
     */
    mapPropertiesUp(array, mapObj) {
        // TODO: Cleaning up old properties isn't working for some reason, it deletes the wrong property even though the key is correct
        // console.warn('mapProperties', array, mapObj);
        return JSON.parse(JSON.stringify(array)).map((element) => {
            // console.warn(element);
            for (const /** @type {?} */ key in mapObj) {
                if (mapObj.hasOwnProperty(key)) {
                    // If the default property needed by the DT is NOT found in the object
                    // This avoids issues where the correct property is actually being passed in even though it is mapped
                    // if (!element[mapObj[key]]) {
                    element[mapObj[key]] = element[key];
                    // console.warn(element);
                    // }
                    // delete element[key];
                    // Possible related to mapping a property down
                }
            }
            return element;
        });
    }
    /**
     * Map the columns into an object based on its property name
     * @param {?} columns
     * @return {?}
     */
    mapColumns(columns) {
        const /** @type {?} */ columnMap = {};
        columns.forEach(column => (columnMap[column.prop] = column));
        return columnMap;
    }
    /**
     * Get the rows that should be visible in the scroll port based on the vertical scroll position
     * @param {?} rows
     * @param {?} scrollProps
     * @param {?} rowsVisible
     * @param {?} rowHeight
     * @return {?}
     */
    getVisibleRows(rows, scrollProps, rowsVisible, rowHeight) {
        // console.log('getVisibleRowsoffSetRowsFromTop', rows, this.scrollProps, this.rowHeight, this.gridProps);
        const /** @type {?} */ rowsNew = [...rows];
        let /** @type {?} */ buffer = 5;
        if (window.navigator.userAgent.indexOf('Edge') > -1) {
            buffer = 15;
        }
        let /** @type {?} */ offSetRowsFromTop = Math.floor(scrollProps.scrollTop / (rowHeight + 1));
        if (offSetRowsFromTop - buffer > 0) {
            offSetRowsFromTop -= buffer;
        }
        if (offSetRowsFromTop < buffer) {
            offSetRowsFromTop = 0;
        }
        let /** @type {?} */ rowsEnd = offSetRowsFromTop + rowsVisible + buffer * 2;
        if (rowsEnd > rowsNew.length) {
            rowsEnd = rowsNew.length;
        }
        return rowsNew.slice(offSetRowsFromTop, rowsEnd);
    }
    /**
     * Create an object of columns that should be visible based on horizontal scroll width
     * @param {?} columns
     * @param {?} scrollProps
     * @param {?} gridProps
     * @return {?}
     */
    getVisibleColumns(columns, scrollProps, gridProps) {
        const /** @type {?} */ colsExternal = [];
        const /** @type {?} */ buffer = 150;
        // let widthTotal = this.scrollProps.scrollLeft;
        let /** @type {?} */ widthCurrent = 0;
        // Loop through column widths
        for (let /** @type {?} */ i = 0; i < columns.length; i++) {
            const /** @type {?} */ column = columns[i];
            // If current column width + all widths before this one is greater than the left scroll position
            // If total column widths is less than the width of the body minus the left scroll position
            if (column.$$width + widthCurrent + buffer > scrollProps.scrollLeft &&
                widthCurrent < gridProps.widthViewPort + scrollProps.scrollLeft + buffer) {
                colsExternal.push(column);
            }
            // Update current width by adding the current column
            widthCurrent = widthCurrent + column.$$width;
        }
        return [...colsExternal];
        // this.columnsExternal = colsExternal;
    }
    /**
     * Filters based on the global object. Also creates the necessary filter structure.
     * @param {?} array
     * @param {?} filterGlobal
     * @return {?}
     */
    filterGlobal(array, filterGlobal) {
        // Loop through the existing props supplied and create the appropriate filter object
        const /** @type {?} */ filters = [];
        filterGlobal.props.forEach(prop => {
            filters.push({
                operator: 'search',
                prop: prop,
                value: filterGlobal.term,
            });
        });
        return this.filterArray(array, filters);
    }
    /**
     * Filter an array of objects with an array of filters
     * @param {?} array
     * @param {?} filters
     * @return {?}
     */
    filterArray(array, filters) {
        // console.warn('filterArray 1: ', array, filters );
        // Get number of contains filters
        const /** @type {?} */ contains = filters.filter(filter => filter.operator === 'contains');
        // Get searches which as passed from the global search option
        const /** @type {?} */ searches = filters.filter(filter => filter.operator === 'search');
        // Equals filters are an OR within the same field but an AND between different fields
        // Map down to dictionary with field as key so we can handle that difference
        const /** @type {?} */ equals = {};
        filters.filter(filter => filter.operator === 'eq').forEach(filter => {
            if (!equals[filter.prop]) {
                equals[filter.prop] = [];
            }
            equals[filter.prop].push(filter);
        });
        // Perform filter and return
        return array.filter(row => {
            // Searches is an OR. If it does not contain, it should be filtered out
            if (searches.length) {
                let /** @type {?} */ matches = false;
                for (let /** @type {?} */ i = 0; i < searches.length; i++) {
                    const /** @type {?} */ filter = searches[i];
                    // If this field isn't found on the row, return false
                    if (!row[filter.prop]) {
                        return false;
                    }
                    // Perform some cleanup on the strings
                    const /** @type {?} */ rowValue = row[filter.prop]
                        .toString()
                        .toLowerCase()
                        .trim();
                    const /** @type {?} */ filterValue = filter.value
                        .toString()
                        .toLowerCase()
                        .trim();
                    if (rowValue.indexOf(filterValue) !== -1) {
                        matches = true;
                        break;
                    }
                }
                // If this group does not have at least one match, return false
                if (!matches) {
                    return false;
                }
            }
            // Contains is an AND. If it does not contain, it should be filtered out
            if (contains.length) {
                let /** @type {?} */ matches = false;
                for (let /** @type {?} */ i = 0; i < contains.length; i++) {
                    const /** @type {?} */ filter = contains[i];
                    // If this field isn't found on the row, return false
                    if (!row[filter.prop]) {
                        return false;
                    }
                    // console.warn(row[filter.prop], '-', filter.prop)
                    // Perform some cleanup on the strings
                    const /** @type {?} */ rowValue = row[filter.prop]
                        .toString()
                        .toLowerCase()
                        .trim();
                    const /** @type {?} */ filterValue = filter.value
                        .toString()
                        .toLowerCase()
                        .trim();
                    if (rowValue.indexOf(filterValue) !== -1) {
                        matches = true;
                        break;
                    }
                }
                // If this group does not have at least one match, return false
                if (!matches) {
                    return false;
                }
            }
            // Equals contains sets of rules groups. Each rule group is based on field and is an OR. Between rule groups it must be an AND
            // Results must contain at least one match within each field
            if (Object.keys(equals).length) {
                let /** @type {?} */ allGroupMatches = 0;
                // Loop through all equals collections
                for (const /** @type {?} */ key in equals) {
                    if (equals.hasOwnProperty(key)) {
                        // Within each field, only need 1 match to include in the collection
                        let /** @type {?} */ matches = false;
                        for (let /** @type {?} */ i = 0; i < equals[key].length; i++) {
                            const /** @type {?} */ filter = equals[key][i];
                            // If this field isn't found on the row, return false
                            if (!row[filter.prop] && row[filter.prop] !== false) {
                                return false;
                            }
                            // If row property is an array of strings, see if the field value is in the array
                            if (Array.isArray(row[filter.prop]) &&
                                row[filter.prop] &&
                                row[filter.prop].indexOf(filter.value.toString()) > -1) {
                                matches = true;
                                break;
                            }
                            // If the field value matches the filter value
                            if (row[filter.prop] === filter.value) {
                                matches = true;
                                break;
                            }
                            // If this is a boolean and needs to match true false
                            if ((filter.value === 'True' && row[filter.prop] === true) ||
                                (filter.value === 'False' && row[filter.prop] === false)) {
                                matches = true;
                                break;
                            }
                        }
                        // If this group has a match, increment the group matcher
                        if (matches) {
                            allGroupMatches++;
                        }
                    }
                }
                if (allGroupMatches !== Object.keys(equals).length) {
                    return false;
                }
            }
            // If no filters failed, return true
            return true;
        });
    }
    /**
     * Sort an array based on a property and direction
     * @param {?} array
     * @param {?} prop
     * @param {?} sortType
     * @return {?}
     */
    sortArray(array, prop, sortType) {
        // console.warn('sortRows', prop, sortType);
        const /** @type {?} */ mapProp = (prop2) => {
            // If string, make lower case and remove special characters
            if (typeof prop2 === 'string') {
                return prop2.toLowerCase().replace(/[^a-z0-9]/gi, '');
            }
            else if (typeof prop2 === 'number') {
                // If number
                return prop2 * 1000;
            }
            else {
                // Make string
                if (prop2) {
                    return prop2.toString();
                }
                return '';
            }
        };
        if (sortType === 'asc') {
            return array.sort((a, b) => (mapProp(a[prop]) !== mapProp(b[prop]) ? (mapProp(a[prop]) < mapProp(b[prop]) ? -1 : 1) : 0));
        }
        else {
            return array.sort((b, a) => (mapProp(a[prop]) !== mapProp(b[prop]) ? (mapProp(a[prop]) < mapProp(b[prop]) ? -1 : 1) : 0));
        }
    }
    /**
     * Group rows by property. Grouping is essentially a multilevel sort
     * @param {?} rows
     * @param {?} columns
     * @param {?} groups
     * @param {?} sorts
     * @param {?} options
     * @return {?}
     */
    groupRows(rows, columns, groups, sorts, options) {
        // console.log('groupRows', groups, sorts);
        const /** @type {?} */ newGroups = {};
        const /** @type {?} */ group = groups[0];
        rows.forEach(row => {
            // If the row property is an array of strings, flatten it down to a string. Otherwise just use string
            let /** @type {?} */ newProp = Array.isArray(row[group.prop]) ? row[group.prop].join() : row[group.prop];
            // If this is an empty array, set to null instead
            // if (Array.isArray(row[group.prop]) && !row[group.prop].length){
            //    newProp = row[group.prop] = null;
            // }
            if (!newProp || newProp === '') {
                newProp = 'No Value';
            }
            // Get current column
            const /** @type {?} */ column = columns.filter(columnNew => columnNew.prop === group.prop);
            if (!newGroups[newProp]) {
                newGroups[newProp] = /** @type {?} */ ({
                    rows: [],
                    column: column && column[0] ? column[0] : null,
                    columnProp: group.prop,
                    columnLabel: '',
                    label: row[group.prop] || 'No Value',
                    visible: true,
                    type: 'group',
                });
                // let label = columns.filter(column => column.prop == group.prop);
                newGroups[newProp].columnLabel = column && column[0] ? column[0].label : 'No Value';
            }
            newGroups[newProp].rows.push(row);
        });
        // Now remap the object into an array
        let /** @type {?} */ grouped = [];
        for (const /** @type {?} */ key in newGroups) {
            if (newGroups.hasOwnProperty(key)) {
                grouped.push(newGroups[key]);
            }
        }
        // Sort the group
        grouped = this.sortArray(grouped, 'label', group.dir);
        let /** @type {?} */ newRows = [];
        const /** @type {?} */ groupsFinal = {};
        // Sort the rows within the group
        grouped.forEach((group2) => {
            if (sorts.length) {
                this.sortArray(group2.rows, sorts[0].prop, sorts[0].dir);
            }
            if (options.primaryKey) {
                // Create a primary key used for the trackRows method.
                //           Group headers are treated as a row and need to have the same primary key as the rest
                (/** @type {?} */ (group2))[options.primaryKey] = group2.label + '-' + group2.rows.length;
            }
            const /** @type {?} */ currentLoc = newRows.length;
            groupsFinal[currentLoc] = group2;
            newRows = [...newRows, group2, ...group2.rows];
        });
        return { rows: newRows, groups: groupsFinal };
    }
    /**
     * Create the statuses/state of the controls (filtering/grouping/sorting)
     * @param {?} state
     * @param {?} columns
     * @return {?}
     */
    createStatuses(state, columns) {
        // console.warn('createStatuses: ',JSON.parse(JSON.stringify(state)));
        const /** @type {?} */ status = {
            groups: {},
            sorts: {},
            filters: {},
        };
        // If groupings are found, create dictionary
        if (state.groups.length) {
            const /** @type {?} */ newStatus = {};
            state.groups.forEach(group => {
                newStatus[group.prop] = group.dir;
            });
            status.groups = newStatus;
        }
        else {
            status.groups = {};
        }
        // If sortings are found, create dictionary
        if (state.sorts.length) {
            const /** @type {?} */ newStatus = {};
            state.sorts.forEach(sort => {
                newStatus[sort.prop] = sort.dir;
            });
            status.sorts = newStatus;
        }
        else {
            status.sorts = {};
        }
        const /** @type {?} */ newFilters = {};
        if (columns && columns.length) {
            columns.forEach(column => {
                newFilters[column.prop] = {
                    contains: '',
                    eq: {},
                };
            });
        }
        // If filters are found, create dictionary
        if (state.filters && typeof state.filters !== 'string' && state.filters.length) {
            state.filters.forEach(filter => {
                // Create field/property object
                if (!newFilters[filter.prop]) {
                    newFilters[filter.prop] = {
                        hasFilters: false,
                        contains: {},
                        eq: {},
                    };
                }
                // Operator is Contains
                if (filter.operator === 'contains') {
                    newFilters[filter.prop][filter.operator] = filter.value;
                    newFilters[filter.prop].hasFilters = true;
                }
                else {
                    // Everything else
                    // Create operator object inside field object
                    if (!newFilters[filter.prop][filter.operator]) {
                        newFilters[filter.prop][filter.operator] = {};
                    }
                    newFilters[filter.prop][filter.operator][filter.value] = filter.value;
                    newFilters[filter.prop].hasFilters = true;
                }
            });
        }
        status.filters = newFilters;
        // console.warn('status 2: ', JSON.parse(JSON.stringify(status)));
        return status;
    }
    /**
     * Look through the rows and assemble a an array of terms
     * @param {?} rows
     * @param {?} columns
     * @return {?}
     */
    getDefaultTermsList(rows, columns) {
        // console.log('getDefaultTermsList');
        // console.time('getDefaultTermsList');
        const /** @type {?} */ termsList = {};
        const /** @type {?} */ uniques = {};
        // Loop through all the columns
        columns.forEach(column => {
            // If the column type is string and does not exist, create the dictionary and array
            if (!termsList[column.prop]) {
                // && column.columnType == 'string'
                termsList[column.prop] = [];
                uniques[column.prop] = {};
            }
        });
        // Find the unique values for each row
        rows.forEach(row => {
            for (const /** @type {?} */ key in termsList) {
                if (termsList.hasOwnProperty(key)) {
                    if ((row[key] || row[key] === false) && uniques[key]) {
                        let /** @type {?} */ keyNew = row[key];
                        // If boolean, convert key to string
                        if (typeof row[key] === 'boolean') {
                            keyNew = startCase(toLower(row[key].toString()));
                        }
                        uniques[key][keyNew] = true;
                    }
                }
            }
        });
        // Now push the uniques to the termslist
        for (const /** @type {?} */ key in uniques) {
            if (uniques.hasOwnProperty(key)) {
                const /** @type {?} */ foo = uniques[key];
                for (const /** @type {?} */ key2 in foo) {
                    if (foo.hasOwnProperty(key2)) {
                        termsList[key].push(key2);
                    }
                }
            }
        }
        // Now sort terms in default order
        for (const /** @type {?} */ key in termsList) {
            if (termsList.hasOwnProperty(key)) {
                // If boolean, have true be first
                if (termsList[key][0] === 'False' || termsList[key][0] === 'True') {
                    termsList[key] = ['True', 'False']; // TODO: Better method of handling boolean than hard coded
                }
                else {
                    termsList[key].sort();
                }
            }
        }
        // console.warn('getDefaultTermsList', termsList);
        // console.timeEnd('getDefaultTermsList');
        return termsList;
    }
    /**
     * Determine the horizontal position of grid cells
     * @param {?} columns
     * @param {?=} offset
     * @return {?}
     */
    columnCalculations(columns, offset = 0) {
        // console.log('columnCalculations', columns, offset);
        let /** @type {?} */ leftOffset = offset;
        return columns.map((column, index) => {
            // If no width, set default to 150
            let /** @type {?} */ width = column.width ? column.width : 150;
            // Ensure min width of 44
            if (width < 44) {
                width = 44;
            }
            // If no width on the column, set a default property
            if (!column.width || !column.$$width) {
                column.width = width;
                column.$$width = width;
            }
            // If no column type, set default of string
            column.columnType = column.columnType ? column.columnType : 'string';
            // Ensure all column widths are divisible by 4, fixes a blurry text bug in chrome
            // column.width = Math.floor(column.width / 4) * 4;
            column.$$index = index;
            column.$$leftOffset = leftOffset;
            leftOffset += width;
            return column;
        });
    }
    /**
     * If total combined width of grid cells is less than viewport, resize widths to match
     * @param {?} columns
     * @param {?} widthColumns
     * @param {?} widthTable
     * @return {?}
     */
    columnsResize(columns, widthColumns, widthTable) {
        let /** @type {?} */ leftOffset = 0;
        const /** @type {?} */ multiplier = Math.floor(widthTable / widthColumns * 100) / 100;
        return columns.map(column => {
            if (column.width) {
                column.$$width = Math.ceil(column.width * multiplier);
                column.$$leftOffset = leftOffset;
                leftOffset += column.$$width;
            }
            return Object.assign({}, column);
        });
    }
    /**
     * Attach any templates to their respective columns
     * Not using map to update columns array because that would retrigger the column getter logic
     * @param {?} columns
     * @param {?} columnTemplates
     * @return {?}
     */
    templatesAddToColumns(columns, columnTemplates) {
        // Loop through supplied columns, attach templates
        const /** @type {?} */ templates = {};
        columns.forEach(column => {
            if (columnTemplates[column.prop]) {
                if (!templates[column.prop]) {
                    templates[column.prop] = {};
                }
                // Cell Templates
                if (columnTemplates[column.prop].templateCell) {
                    templates[column.prop].templateCell = columnTemplates[column.prop].templateCell;
                }
                // Header Templates
                if (columnTemplates[column.prop].templateCell) {
                    templates[column.prop].templateHeader = columnTemplates[column.prop].templateHeader;
                }
            }
        });
        return templates;
    }
    /**
     * Map custom templates to a usable object for references
     * @param {?} arr
     * @return {?}
     */
    templatesMapColumns(arr) {
        const /** @type {?} */ result = {};
        for (const /** @type {?} */ temp of arr) {
            const /** @type {?} */ col = {};
            const /** @type {?} */ props = Object.getOwnPropertyNames(temp);
            for (const /** @type {?} */ prop of props) {
                col[prop] = temp[prop];
            }
            if (temp.headerTemplate) {
                col.templateHeader = temp.templateHeader;
            }
            if (temp.cellTemplate) {
                col.templateCell = temp.templateCell;
            }
            result[col.prop] = col;
        }
        return result;
    }
    /**
     * Calculate row properties such as visibility, y position and index
     * @param {?} rows
     * @param {?} rowHeight
     * @param {?=} makeVisible
     * @return {?}
     */
    rowPositions(rows, rowHeight, makeVisible = false) {
        let /** @type {?} */ y = 0;
        return rows.filter((row, i) => {
            row.$$rowIndex = i; // Set rowIndex
            // If hidden prop is not set, set default to false
            if (typeof row.$$hidden === 'undefined' || makeVisible) {
                row.$$hidden = false;
            }
            // If visible
            if (row.$$hidden === false) {
                row.$$rowPosition = y; // Set y position
                y += rowHeight + 1;
                return true;
            }
            else {
                return false;
            }
        });
    }
}
DataGridService.decorators = [
    { type: Injectable },
];
/** @nocollapse */
DataGridService.ctorParameters = () => [];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class DataGridColumnHeaderDirective {
    /**
     * @param {?} template
     */
    constructor(template) {
        this.template = template;
    }
}
DataGridColumnHeaderDirective.decorators = [
    { type: Directive, args: [{ selector: '[datagrid-header-template]' },] },
];
/** @nocollapse */
DataGridColumnHeaderDirective.ctorParameters = () => [
    { type: TemplateRef, },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class DataGridColumnCellDirective {
    /**
     * @param {?} template
     */
    constructor(template) {
        this.template = template;
    }
}
DataGridColumnCellDirective.decorators = [
    { type: Directive, args: [{ selector: '[datagrid-cell-template]' },] },
];
/** @nocollapse */
DataGridColumnCellDirective.ctorParameters = () => [
    { type: TemplateRef, },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class DataTableColumnDirective {
}
DataTableColumnDirective.decorators = [
    { type: Directive, args: [{ selector: 'datagrid-column' },] },
];
/** @nocollapse */
DataTableColumnDirective.propDecorators = {
    "name": [{ type: Input },],
    "prop": [{ type: Input },],
    "templateCell": [{ type: Input }, { type: ContentChild, args: [DataGridColumnCellDirective, { read: TemplateRef },] },],
    "templateHeader": [{ type: Input }, { type: ContentChild, args: [DataGridColumnHeaderDirective, { read: TemplateRef },] },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/** @enum {string} */
const Actions = {
    group: 'group',
    sort: 'sort',
    pinLeft: 'pinLeft',
    filter: 'filter',
    reset: 'reset',
    column: 'column',
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
const throttle = require('lodash/throttle');
class BodyComponent {
    /**
     * @param {?} zone
     * @param {?} element
     */
    constructor(zone, element) {
        this.zone = zone;
        this.onColumnsUpdated = new EventEmitter();
        this.onStateUpdated = new EventEmitter();
        this.onCustomLinkEvent = new EventEmitter();
        this.onRowUpdated = new EventEmitter();
        this.onRowMouseEvent = new EventEmitter();
        this.onGroupToggled = new EventEmitter();
        this.onRightClick = new EventEmitter();
        this.onRowMouseDown = new EventEmitter();
        this.onRowMouseUp = new EventEmitter();
        this.onScrollEvent = new EventEmitter();
        /**
         * Throttle the scroll event
         */
        this.onScrollThrottled = throttle((event) => this.onScroll(event), 20, { trailing: true, leading: true });
        this.body = element.nativeElement;
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        // Attach scroll event listener
        this.zone.runOutsideAngular(() => {
            this.body.addEventListener('scroll', this.onScrollThrottled.bind(this), /** @type {?} */ ({ capture: true, passive: true }));
        });
    }
    /**
     * When the datatable is scrolled
     * @param {?} event
     * @return {?}
     */
    onScroll(event) {
        // console.log('onScroll', event.target.scrollTop);
        requestAnimationFrame(() => {
            const /** @type {?} */ scrollProps = {
                scrollTop: event.target.scrollTop,
                scrollLeft: event.target.scrollLeft,
            };
            this.onScrollEvent.emit(scrollProps);
        });
    }
    /**
     * Return a unique ID to ngfor to improve performance
     * @param {?} _index
     * @param {?} item - The column
     * @return {?}
     */
    trackRow(_index, item) {
        // console.log('Tracking Rows');
        return item.$$track;
    }
    /**
     * Communicate mouse events on the body row up to the parent. Used mainly for selection
     * @param {?} type
     * @param {?} rowIndex
     * @param {?=} event
     * @return {?}
     */
    onMouseEvent(type, rowIndex, event) {
        this.onRowMouseEvent.emit({ type: type, rowIndex: rowIndex, event: event });
    }
    /**
     * Pass updated field up the component chain
     * @param {?} event
     * @param {?} rowIndex
     * @return {?}
     */
    rowUpdated(event, rowIndex) {
        event.rowIndex = rowIndex;
        this.onRowUpdated.emit(event);
    }
    /**
     * When a group is toggled
     * @param {?} group
     * @return {?}
     */
    groupToggled(group) {
        this.onGroupToggled.emit(group);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        // Unsub from all subscriptions
        this.body.removeEventListener('scroll', this.onScrollThrottled.bind(this), /** @type {?} */ ({ capture: true, passive: true }));
    }
}
BodyComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid-body',
                template: `<div class="datagrid-body-container" id="dataGridBody" #dataGridBody
	 [ngStyle]="{'height': gridProps.heightTotal + 'px'}"
	 [ngClass]="{ 'widthFixed' : gridProps.widthFixed }">

	<div class="datagrid-body" [ngStyle]="{'width' : gridProps.widthViewPort + 'px', 'height' : gridProps.heightBody + 'px' }">

		<ng-container *ngFor="let row of rows; let rowIndex = index; trackBy:trackRow;">

			<div class="datagrid-body-row" *ngIf="!row.$$hidden"
				 [style.top]="row.$$rowPosition + 'px'"
				 [ngClass]="{'selected': row.$$selected }">

				<!-- Group header. Treated as a row -->
				<group-header *ngIf="row.type == 'group' "
							  [group]="row"
							  [options]="options"
							  [width]="gridProps.widthTotal"
							  (onGroupToggled)="groupToggled($event)">
				</group-header>

				<!-- Normal Row -->
				<ng-container *ngIf="row.type != 'group' ">

					<!-- Pinned columns -->
					<div class="datagrid-body-pinned sticky-left" *ngIf="columnsPinnedLeft?.length"
						 [style.width]="gridProps.widthPinned + 'px'"
						 [ngStyle]="rowStyles[row[options.primaryKey]]"
						 (click)="onMouseEvent('click', row.$$rowIndex, $event)"
						 (dblclick)="onMouseEvent('dblclick', row.$$rowIndex, $event)"
						 (contextmenu)="onMouseEvent('contextmenu', row.$$rowIndex, $event)"
						 (mousedown)="onMouseEvent('mousedown', row.$$rowIndex, $event)"
						 (mouseup)="onMouseEvent('mouseup', row.$$rowIndex, $event)"
						 (mouseenter)="onMouseEvent('mouseenter', row.$$rowIndex, $event)">
						<datagrid-body-row [columns]="columnsPinnedLeft"
										   [row]="row"
										   [options]="options"
                       [templates]="templates"
										   (onRowUpdated)="rowUpdated($event, row.$$rowIndex)">
						</datagrid-body-row>
					</div>

					<!-- Non-pinned columns -->
					<div class="datagrid-body-main" style="display: inline-block"
						 [style.width]="gridProps.widthMain + 'px'"
						 [ngStyle]="rowStyles[row[options.primaryKey]]"
						 (click)="onMouseEvent('click', row.$$rowIndex, $event)"
						 (dblclick)="onMouseEvent('dblclick', row.$$rowIndex, $event)"
						 (contextmenu)="onMouseEvent('contextmenu', row.$$rowIndex, $event)"
						 (mousedown)="onMouseEvent('mousedown', row.$$rowIndex, $event)"
						 (mouseup)="onMouseEvent('mouseup', row.$$rowIndex, $event)"
						 (mouseenter)="onMouseEvent('mouseenter', row.$$rowIndex, $event)">
						<datagrid-body-row [columns]="columns"
										   [row]="row"
										   [options]="options"
                       [templates]="templates"
										   (onRowUpdated)="rowUpdated($event, row.$$rowIndex)">
						</datagrid-body-row>
					</div>

				</ng-container>

			</div>
		</ng-container>
	</div>
</div>
`,
                styles: [],
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
BodyComponent.ctorParameters = () => [
    { type: NgZone, },
    { type: ElementRef, },
];
BodyComponent.propDecorators = {
    "columns": [{ type: Input },],
    "columnsPinnedLeft": [{ type: Input },],
    "rows": [{ type: Input },],
    "rowStyles": [{ type: Input },],
    "state": [{ type: Input },],
    "status": [{ type: Input },],
    "options": [{ type: Input },],
    "gridProps": [{ type: Input },],
    "templates": [{ type: Input },],
    "onColumnsUpdated": [{ type: Output },],
    "onStateUpdated": [{ type: Output },],
    "onCustomLinkEvent": [{ type: Output },],
    "onRowUpdated": [{ type: Output },],
    "onRowMouseEvent": [{ type: Output },],
    "onGroupToggled": [{ type: Output },],
    "onRightClick": [{ type: Output },],
    "onRowMouseDown": [{ type: Output },],
    "onRowMouseUp": [{ type: Output },],
    "onScrollEvent": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
const throttle$1 = require('lodash/throttle');
/**
 * Documentation and scaffolding available in this folder in datatable.scaffold.ts *
 */
class DataGridComponent {
    /**
     * @param {?} dgSvc
     * @param {?} ref
     */
    constructor(dgSvc, ref) {
        this.dgSvc = dgSvc;
        this.ref = ref;
        /**
         * State. Set default if state is not passed down from parent
         */
        this._state = {
            filters: [],
            sorts: [],
            groups: [],
            info: {},
        };
        /**
         * Outputs
         */
        this.onColumnsUpdated = new EventEmitter();
        this.onRowsSelected = new EventEmitter();
        this.onStateChange = new EventEmitter();
        this.onRightClickMenu = new EventEmitter();
        this.action = new EventEmitter();
        this.onCustomLinkEvent = new EventEmitter();
        this.onElementRef = new EventEmitter();
        this.onRowUpdated = new EventEmitter();
        /**
         * Columns that are pinned
         */
        this.columnsPinnedLeft = [];
        /**
         * Columns that are pinned
         */
        this.columnsPinnedRight = [];
        /**
         * Properties and info about the grid itself, IE formatting such as width and scroll area
         */
        this.gridProps = {
            widthTotal: 0,
            widthPinned: 0,
            widthMain: 0,
            heightTotal: 0,
            heightBody: 0,
            widthViewPort: 0,
            widthFixed: true,
        };
        /**
         * Properties related to scrolling of the main grid
         */
        this.scrollProps = { scrollTop: 0, scrollLeft: 0 };
        /**
         * Holds custom templates for cells
         */
        this.templatesCell = {};
        /**
         * A dictionary of columns based on primary key, used for lookups
         */
        this.columnsMapped = {};
        /**
         * A dictionary that holds css CLASSES for a given row.
         * The lookup is the primary key specified in the options. Gets its data from options.rowClass
         */
        this.rowClasses = {};
        /**
         * A dictionary that holds css STYLES for a given row.
         * The lookup is the primary key specified in the options. Gets its data from options.rowStyle
         */
        this.rowStyles = {};
        /**
         * Does the datatable have the data it needs to draw the dom?
         */
        this.appReady = false;
        /**
         * Does the datatable have the data it needs to draw the dom?
         */
        this.domReady = false;
        /**
         * Is the user dragging with the mouse
         */
        this.dragging = false;
        this.draggingPos = {
            hasMinSize: false,
            startX: 0,
            startY: 0,
            top: 0,
            bottom: 0,
            left: 0,
            right: 0,
            width: 0,
            height: 0,
            bounding: {
                top: 0,
                bottom: 0,
                left: 0,
                right: 0,
            },
        };
        /**
         * A dictionary of currently pressed keys
         */
        this.keysPressed = {};
        /**
         * Hold the rowindex and group index when a row is clicked and dragged
         */
        this.rowClickDrag = {
            rowIndex: 0,
            groupIndex: 0,
        };
        /**
         * The sum of the current column widths. Used to determine if column resize is necessary
         */
        this.columnWidthsInternal = 0;
        /**
         * The height of the row. Necessary for virtual scroll calculation.
         * Needs to be an odd number to prevent partial pixel problems. Has 1px border added
         */
        this.rowHeight = 23;
        /**
         * Keep track of which indexes are visible to prevent the component tree
         * from being updated unless actually changed
         */
        this.rowsIndexes = { start: 0, end: 0 };
        this.columnIndexes = { start: 0, end: 0 };
        /**
         * Throttle the window resize event
         */
        this.onWindowResizeThrottled = throttle$1(() => this.onWindowResize(), 300, { trailing: true, leading: true });
        /**
         * Hold subs for future unsub
         */
        this.subscriptions = [];
        this.rowGroups = [];
        this.groups = {};
    }
    /**
     * @param {?} columns
     * @return {?}
     */
    set columns(columns) {
        if (columns && columns.length) {
            // Create a random number slug so if different columns are passed a new instance is created every time
            const /** @type {?} */ slug = Math.floor(Math.random() * 1000000);
            // Create custom track property and new reference for each column
            columns = columns.map((column, i) => {
                column.$$track = slug + '-' + i;
                return Object.assign({}, column);
            });
            // If columnMap object is supplied, remap column props to what the datatable needs
            if (this.options && this.options.columnMap) {
                columns = this.dgSvc.mapPropertiesDown(columns, this.options.columnMap);
            }
        }
        this._columns = columns;
    }
    /**
     * @return {?}
     */
    get columns() {
        return this._columns && this._columns.length ? [...this._columns] : [];
    }
    /**
     * @param {?} rows
     * @return {?}
     */
    set rows(rows) {
        if (rows && rows.length) {
            // Create a random number slug so if different rows are passed a new instance is created every time
            const /** @type {?} */ slug = Math.floor(Math.random() * 1000000);
            rows.forEach((row, i) => {
                row.$$track = slug + '-' + i;
                row.$$selected = false;
            }); // Add the unique ID which is slug + index
        }
        this._rows = rows;
    }
    /**
     * @return {?}
     */
    get rows() {
        return this._rows && this._rows.length ? [...this._rows] : [];
    }
    /**
     * @param {?} state
     * @return {?}
     */
    set state(state) {
        // If no state passed down, set a default and empty state object
        const /** @type {?} */ stateNew = state ? state : {};
        if (!stateNew.filters) {
            stateNew.filters = [];
        }
        if (!stateNew.sorts) {
            stateNew.sorts = [];
        }
        if (!stateNew.groups) {
            stateNew.groups = [];
        }
        if (!stateNew.info) {
            stateNew.info = {};
        }
        // If controls map is specified, map state property to appropriate fields
        stateNew.groups =
            stateNew.groups && stateNew.groups.length && this.options.controlsMap
                ? this.dgSvc.mapPropertiesDown(stateNew.groups, this.options.controlsMap)
                : stateNew.groups;
        stateNew.sorts =
            stateNew.sorts && stateNew.sorts.length && this.options.controlsMap
                ? this.dgSvc.mapPropertiesDown(stateNew.sorts, this.options.controlsMap)
                : stateNew.sorts;
        stateNew.filters =
            stateNew.filters && stateNew.filters.length && this.options.controlsMap
                ? this.dgSvc.mapPropertiesDown(stateNew.filters, this.options.controlsMap)
                : stateNew.filters;
        stateNew.info = {
            initial: true,
        };
        this._state = stateNew;
    }
    /**
     * @return {?}
     */
    get state() {
        return this._state;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set columnTemplates(val) {
        const /** @type {?} */ arr = val.toArray();
        if (arr.length) {
            this._columnTemplates = this.dgSvc.templatesMapColumns(arr);
        }
    }
    /**
     * @return {?}
     */
    get columnTemplates() {
        return this._columnTemplates;
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * @param {?} model
     * @return {?}
     */
    ngOnChanges(model) {
        // Clear all memoized caches anytime new data is loaded into the grid
        // this.dgSvc.clearCaches();
        // If filter global is set or updated
        if (model.filterGlobal && this.filterGlobal && this.filterGlobal.term) {
            throttle$1(() => this.viewCreate(), 500, { trailing: true, leading: true });
        }
        // If NEW columns are passed
        if (model.columns) {
            // If columnMap object is supplied, remap column props to what the datatable needs
            const /** @type {?} */ columns = this.options.columnMap
                ? this.dgSvc.mapPropertiesDown(this.columns, this.options.columnMap)
                : this.columns;
            const /** @type {?} */ columnsPinnedLeft = columns.filter(column => (column.pinnedLeft ? true : false));
            this.columnsPinnedLeft = columns.length ? this.dgSvc.columnCalculations(columnsPinnedLeft) : [];
            // Get un-pinned columns
            const /** @type {?} */ columnsInternal = columns.filter(column => (!column.pinnedLeft ? true : false));
            this.columnsInternal = columns.length ? this.dgSvc.columnCalculations(columnsInternal) : [];
            // Determine total width of internal columns
            this.columnWidthsInternal = this.columnsInternal.reduce((a, b) => (b.width ? a + b.width : 0), 0);
            // Create a column map
            this.columnsMapped = this.dgSvc.mapColumns(this.columns);
        }
        // If state and columns are present, check to make sure state has valid fields
        if (this.state && this.columns) {
            // Loop through each element in the state object and verify that the columns exist
            // Delete columns that are referenced in state that don't exist. Helps protect against corrupt states
            if (this.state.sorts &&
                this.state.sorts.length &&
                this.state.sorts[0].prop &&
                !this.columnsMapped[this.state.sorts[0].prop]) {
                this.state.sorts = [];
                console.error(`Sorting option is for a column that doesn't exist. Sort option has been removed.`);
            }
            if (this.state.groups && this.state.groups.length && !this.columnsMapped[this.state.groups[0].prop]) {
                this.state.groups = [];
                console.error(`Grouping option is for a column that doesn't exist. Group option has been removed.`);
            }
            if (this.state.filters && this.state.filters.length) {
                for (let /** @type {?} */ i = this.state.filters.length - 1; i >= 0; i--) {
                    if (!this.columnsMapped[this.state.filters[i].prop]) {
                        this.state.filters = this.state.filters.filter((_filter, index) => i !== index);
                        console.error(`Filter option is for a column that doesn't exist. Filter option has been removed.`);
                    }
                }
            }
        }
        // If columns and rows are available
        if (this.columns && this.rows) {
            // On any column or row changes, unselect rows
            this.rows.forEach(row => (row.$$selected = false));
            this.filterTerms = this.dgSvc.getDefaultTermsList(this.rows, this.columns); // Generate a list of default filter terms
            this.createRowStyles(); // Create row styles
            this.createRowClasses(); // Create row classes
            // Only on initial load, set app ready. This prevents the app from hanging on a route change
            this.appReady = true;
            this.dataGridReady();
            // Emit the state change to the parent component now that the first initial view has been created
            this.emitState(this.state);
        }
    }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        this.domReady = true;
        this.dataGridReady();
    }
    /**
     * Determine the conditions for when the datagrid is ready to render to the dom
     * TODO: Set max iterations to check to avoid infinite loop
     * @return {?}
     */
    dataGridReady() {
        // If appdata and DOM are ready
        if (this.appReady && this.domReady) {
            // Make sure that the datagrid is visible on the DOM so that the width can be extracted
            if (this.dataGrid && this.dataGrid.nativeElement) {
                this.gridProps.widthViewPort = Math.floor(this.dataGrid.nativeElement.getBoundingClientRect().width);
            }
            // If width is available, create the view and render the dom
            if (this.gridProps.widthViewPort) {
                this.viewCreate();
                // this.ref.detectChanges();
            }
            else {
                // If for some reason the DOM is not available, check every 100 seconds until it is
                setTimeout(() => {
                    this.dataGridReady();
                }, 50);
            }
        }
    }
    /**
     * When the datatable is scrolled
     * @param {?} scrollPropsNew
     * @return {?}
     */
    onScroll(scrollPropsNew) {
        // console.log('onScroll', scrollPropsNew);
        this.ref.detach();
        const /** @type {?} */ scrollPropsOld = Object.assign({}, this.scrollProps);
        const /** @type {?} */ scrollProps = Object.assign({}, scrollPropsNew);
        // Update rows only if rows have changed
        if (scrollPropsOld.scrollTop !== scrollProps.scrollTop) {
            const /** @type {?} */ rowsVisible = Math.ceil(this.gridProps.heightTotal / this.rowHeight);
            const /** @type {?} */ rowsExternal = this.dgSvc.getVisibleRows(this.rowsInternal, scrollProps, rowsVisible, this.rowHeight);
            if (!this.rowsIndexes ||
                (rowsExternal[0] &&
                    rowsExternal[0].$$track !== this.rowsIndexes.start &&
                    rowsExternal[rowsExternal.length - 1].$$track !== this.rowsIndexes.end)) {
                this.rowsExternal = rowsExternal;
                this.rowsIndexes = {
                    start: rowsExternal[0].$$track,
                    end: rowsExternal[rowsExternal.length - 1].$$track,
                };
            }
        }
        // Update columns only if columns have changed
        if (scrollPropsOld.scrollLeft !== scrollProps.scrollLeft) {
            const /** @type {?} */ columnsExternal = this.dgSvc.getVisibleColumns(this.columnsInternal, scrollProps, this.gridProps);
            if (!this.columnIndexes ||
                (columnsExternal[0] &&
                    columnsExternal[0].$$track !== this.columnIndexes.start &&
                    columnsExternal[columnsExternal.length - 1].$$track !== this.columnIndexes.end)) {
                this.columnsExternal = columnsExternal;
                this.columnIndexes = {
                    start: columnsExternal[0].$$track,
                    end: columnsExternal[columnsExternal.length - 1].$$track,
                };
            }
        }
        this.scrollProps = scrollProps;
        this.ref.reattach();
        // Does not scroll seamlessly without calling detectChanges
        // TODO: Detect changes is pretty costly, need another way to determine change detection
        this.ref.detectChanges();
    }
    /**
     * Create the view by assembling everything that modifies the state
     * @return {?}
     */
    viewCreate() {
        // console.warn('createView', this.state, this.status, this.filterTerms );
        // console.time('Creating View');
        // Set manual change detection
        this.ref.detach();
        let /** @type {?} */ newRows = this.rows;
        // console.log('Total Rows', newRows.length)
        // If global filter option is set filter
        if (this.filterGlobal && this.filterGlobal.term) {
            // TODO: Figure out how to memoize since groupRows is not a pure function
            newRows = this.dgSvc.filterGlobal(newRows, this.filterGlobal);
        }
        // If custom filters are specified
        if (this.state && this.state.filters && this.state.filters.length) {
            newRows = this.dgSvc.filterArray(newRows, this.state.filters);
        }
        // If grouped
        if (this.state && this.state.groups && this.state.sorts && this.state.groups.length) {
            // Create groups
            // TODO: Figure out how to memoize since groupRows is not a pure function
            const /** @type {?} */ groupings = this.dgSvc.groupRows(newRows, this.columns, this.state.groups, this.state.sorts, this.options);
            newRows = groupings.rows;
            this.groups = groupings.groups;
        }
        else {
            // If NOT grouped
            this.groups = null;
            // If sorts
            if (this.state &&
                this.state.sorts &&
                this.state.sorts.length &&
                this.state.sorts[0] &&
                this.state.sorts[0].prop &&
                this.state.sorts !== undefined) {
                // Sort Arrays
                newRows = this.dgSvc.sortArray(newRows, this.state.sorts[0].prop, this.state.sorts[0].dir);
            }
        }
        // Generate row vertical positions
        newRows = this.dgSvc.rowPositions(newRows, this.rowHeight);
        this.updateGridProps();
        // Set updated columns
        this.columnsPinnedLeft = this.columnsPinnedLeft.length ? this.dgSvc.columnCalculations(this.columnsPinnedLeft) : [];
        this.columnsInternal = this.columnsInternal.length ? this.dgSvc.columnCalculations(this.columnsInternal) : [];
        // If the total width of the columns is less than the viewport, resize columns to fit
        if (this.dataGrid &&
            this.dataGrid.nativeElement &&
            this.columnWidthsInternal < this.dataGrid.nativeElement.getBoundingClientRect().width) {
            // Resize columns to fit available space
            this.columnsInternal = this.dgSvc.columnsResize(this.columnsInternal, this.columnWidthsInternal, this.gridProps.widthViewPort - this.gridProps.widthPinned);
            this.gridProps.widthFixed = true;
        }
        else {
            // Reset widths
            this.columnsInternal = this.columnsInternal.map(column => {
                column.$$width = column.width;
                return column;
            });
            this.gridProps.widthFixed = false;
        }
        this.updateGridProps();
        // Update internal modified rows
        this.rowsInternal = newRows;
        // Update columns to go to the DOM
        this.columnsExternal = this.dgSvc.getVisibleColumns(this.columnsInternal, this.scrollProps, this.gridProps);
        // Updated rows to go to the DOM
        const /** @type {?} */ rowsVisible = Math.ceil(this.gridProps.heightTotal / this.rowHeight);
        this.rowsExternal = this.dgSvc.getVisibleRows(this.rowsInternal, this.scrollProps, rowsVisible, this.rowHeight);
        // this.rowsExternal = this.dgSvc.getVisibleRows(this.rowsInternal, this.scrollProps, rowsVisible, this.rowHeight);
        if (this.state.info) {
            // Add stats and info to be emitted
            this.state.info.rowsTotal = this.rows.length;
            this.state.info.rowsVisible = this.rowsInternal.filter(row => !row.type).length; // Filter out any group columns
        }
        this.status = this.dgSvc.createStatuses(this.state, this.columnsInternal);
        // console.warn('this.status', this.status);
        this.state = Object.assign({}, this.state);
        // Emit the state change to the parent component
        this.emitState(this.state);
        // Turn change detection back on
        this.ref.reattach();
        // console.timeEnd('Creating View');
    }
    /**
     * When a group is toggled
     * @return {?}
     */
    groupToggled() {
        this.viewCreate();
    }
    /**
     * When the datatable state is changed, usually via a control such as group/filter/sort etc
     * @param {?} stateChange
     * @return {?}
     */
    onStateUpdated(stateChange) {
        // console.warn('changeState ', stateChange);
        this.ref.detach();
        const /** @type {?} */ newState = Object.assign({}, this.state);
        // Legacy support for previous states of the grid. Ensure all arrays exist to prevent errors
        if (!newState.filters || !Array.isArray(newState.filters)) {
            newState.filters = [];
        }
        if (!newState.sorts) {
            newState.sorts = [];
        }
        if (!newState.groups) {
            newState.groups = [];
        }
        if (!newState.info) {
            newState.info = {};
        }
        newState.info.initial = false;
        // ### Update Sorting ###
        if (stateChange.action === Actions.sort) {
            newState.sorts = stateChange.data.dir ? [stateChange.data] : [];
        }
        else if (stateChange.action === Actions.group) {
            // ### Update Grouping ###
            newState.groups = stateChange.data.dir ? [stateChange.data] : [];
        }
        else if (stateChange.action === Actions.filter) {
            // ### Update Filtering ###
            const /** @type {?} */ newFilter = stateChange.data.filter;
            if (stateChange.data.filterAction === 'change') {
                let /** @type {?} */ index = 0;
                for (let /** @type {?} */ i = 0; i < newState.filters.length; i++) {
                    if (newFilter.prop === newState.filters[i].prop && newFilter.operator === newState.filters[i].operator) {
                        index = i;
                        break;
                    }
                }
                newState.filters[index] = newFilter;
            }
            else if (stateChange.data.filterAction === 'add') {
                newState.filters.push(newFilter);
            }
            else if (stateChange.data.filterAction === 'remove') {
                let /** @type {?} */ index = 0;
                for (let /** @type {?} */ i = 0; i < newState.filters.length; i++) {
                    // If this is a contains filter, only match against field and operator
                    if (newFilter.operator === 'contains' &&
                        newFilter.prop === newState.filters[i].prop &&
                        newFilter.operator === newState.filters[i].operator) {
                        index = i;
                        break;
                    }
                    else if (newFilter.prop === newState.filters[i].prop &&
                        newFilter.operator === newState.filters[i].operator &&
                        newFilter.value === newState.filters[i].value) {
                        // If not a contains filter, match against all 3 fields
                        index = i;
                        break;
                    }
                }
                newState.filters.splice(index, 1);
            }
            else if (stateChange.data.filterAction === 'clear') {
                newState.filters = newState.filters.filter(item => item.prop !== stateChange.data.filter.prop);
                // console.warn('Clearing filters', newState.filters);
            }
        }
        else if (stateChange.action === Actions.reset) {
            // ### Reset everything ###
            // newRows = [...this.rows];
        }
        else if (stateChange.action === Actions.column) {
            // ### Column Changes ###
            // Deletion
            if (stateChange.data.action === 'delete') {
                this.columnsInternal = this.columnsInternal.filter(column => column.$$index !== stateChange.data.columnIndex);
                this.columns = this.columns.filter(column => column.$$index !== stateChange.data.columnIndex);
                // Update total width of internal columns
                this.columnWidthsInternal = this.columnsInternal.reduce((a, b) => (b.width ? a + b.width : 0), 0);
                this.emitColumns();
            }
        }
        else if (stateChange.action === Actions.pinLeft) {
            // ### Pinning ###
            if (stateChange.data.isPinned) {
                // Get column being unpinned
                const /** @type {?} */ colNew = this.columnsPinnedLeft[stateChange.data.index];
                delete colNew.pinnedLeft; // Delete pinned prop
                // Remove from pinned array
                this.columnsPinnedLeft = this.columnsPinnedLeft.filter((_col, index) => stateChange.data.index !== index);
                // Add to main array
                this.columnsInternal = [colNew, ...this.columnsInternal];
            }
            else {
                // console.warn('Pinning to left', stateChange.data);
                // Get pinned column
                const /** @type {?} */ newCol = this.columnsInternal.filter(col => col.prop === stateChange.data.prop)[0];
                newCol.pinnedLeft = true;
                this.columnsPinnedLeft = [...this.columnsPinnedLeft, newCol];
                // Update non pinned columns
                this.columnsInternal = this.columnsInternal.filter(col => col.prop !== stateChange.data.prop);
            }
            // Update total width of internal columns
            this.columnWidthsInternal = this.columnsInternal.reduce((a, b) => (b.width ? a + b.width : 0), 0);
            this.emitColumns();
        }
        this.state = newState;
        // Now create the view and update the DOM
        this.viewCreate();
    }
    /**
     * When columns are modified from a lower component
     * @param {?} columnData
     * @return {?}
     */
    columnsUpdated(columnData) {
        // console.log('columnsUpdated', columnData);
        // If this is a resize column event
        if (columnData && columnData.columnIndex) {
            if (columnData.action === 'resize') {
                // Determine if updating pinned or regular columns
                if (columnData.type === 'pinnedLeft') {
                    this.columnsPinnedLeft[columnData.columnIndex].width = columnData.width;
                    this.columnsPinnedLeft[columnData.columnIndex] = Object.assign({}, this.columnsPinnedLeft[columnData.columnIndex]);
                    // this.columnsPinnedLeft = [...this.columnsPinnedLeft];
                }
                else {
                    this.columnsInternal[columnData.columnIndex].width = columnData.width;
                    this.columnsInternal[columnData.columnIndex] = Object.assign({}, this.columnsInternal[columnData.columnIndex]);
                    // this.columnsInternal = [...this.columnsInternal];
                }
            }
            else if (columnData.action === 'reorder') {
                // If this is a reorder columns event
                if (columnData.type === 'pinnedLeft') {
                    const /** @type {?} */ colOld = this.columnsPinnedLeft.filter(column => column.prop === columnData.prop)[0]; // Get column being moved
                    const /** @type {?} */ colsNew = this.columnsPinnedLeft.filter(column => column.prop !== columnData.prop); // Get new array without that column
                    colsNew.splice(columnData.columnIndex, 0, colOld); // Insert into index location
                    this.columnsPinnedLeft = colsNew; // Update reference
                }
                else {
                    const /** @type {?} */ colOld = this.columnsInternal.filter(column => column.prop === columnData.prop)[0]; // Get column being moved
                    const /** @type {?} */ colsNew = this.columnsInternal.filter(column => column.prop !== columnData.prop); // Get new array without that column
                    colsNew.splice(columnData.columnIndex, 0, colOld); // Insert into index location
                    this.columnsInternal = colsNew; // Update reference
                }
            }
            // Update total width of internal columns
            this.columnWidthsInternal = this.columnsInternal.reduce((a, b) => (b.width ? a + b.width : 0), 0);
            this.emitColumns();
            this.viewCreate();
        }
    }
    /**
     * Global properties needed by grid to draw itself
     * @return {?}
     */
    updateGridProps() {
        const /** @type {?} */ gridProps = Object.assign({}, this.gridProps);
        // Get width of DOM viewport
        if (this.dataGrid && this.dataGrid.nativeElement.getBoundingClientRect().width) {
            gridProps.widthViewPort = Math.floor(this.dataGrid.nativeElement.getBoundingClientRect().width) || 0;
        }
        // Get width of pinned columns
        gridProps.widthPinned = this.columnsPinnedLeft.length
            ? this.columnsPinnedLeft.reduce((a, b) => (b.$$width ? a + b.$$width : 0), 0)
            : 0;
        // Get width of non-pinned columns
        gridProps.widthMain = this.columnsInternal.reduce((a, b) => (b.$$width ? a + b.$$width : 0), 0) || 0;
        // Get width of internal columns plus pinned columns
        gridProps.widthTotal = gridProps.widthMain + gridProps.widthPinned;
        // Get height of grid
        if (this.options.heightMax) {
            gridProps.heightTotal = /** @type {?} */ (this.options.heightMax);
        }
        else if (this.options.fullScreen) {
            const /** @type {?} */ height = this.dataGrid.nativeElement.getBoundingClientRect().height;
            let /** @type {?} */ newHeight = height - 2 - this.rowHeight; // Add offsets for table header and bottom scrollbar
            // Check if the info bar is showing, deduct from total height
            if (this.options.showInfo && (this.state.sorts.length || this.state.groups.length || this.state.filters.length)) {
                newHeight -= this.rowHeight;
            }
            gridProps.heightTotal = newHeight;
        }
        else {
            // Set default height if non specified
            gridProps.heightTotal = 300;
        }
        // gridProps.rowsVisible = Math.ceil(gridProps.heightTotal / this.rowHeight); // Get max visible rows
        if (this.rowsInternal && this.rowsInternal.length) {
            gridProps.heightBody = this.rowsInternal.length * this.rowHeight;
        }
        else if (this.rows && this.rows.length) {
            gridProps.heightBody = this.rows.length * this.rowHeight;
        }
        else {
            gridProps.heightBody = 300;
        }
        this.gridProps = gridProps;
    }
    /**
     * On a global mouse down event
     * @param {?} event
     * @return {?}
     */
    handleMouseDown(event) {
        // console.warn('handleMouseDown 1', event, this.dataGridBody)
        // Set the default starting position of the initial click and also get the bounding box of the datatable
        const /** @type {?} */ draggingPos = Object.assign({ startX: event.pageX, startY: event.pageY, bounding: this.dataGridBody.body.getBoundingClientRect() }, this.draggingPos);
        // Only drag on left mouse click
        // Make sure the drag starts within the datatable bounding box
        if (event.which === 1 &&
            draggingPos.startY > draggingPos.bounding.top &&
            draggingPos.startY < draggingPos.bounding.bottom &&
            draggingPos.startX > draggingPos.bounding.left &&
            draggingPos.startX < draggingPos.bounding.right) {
            this.draggingPos = draggingPos;
            this.dragging = true;
        }
    }
    /**
     * Global mouse up event
     * @param {?} event
     * @return {?}
     */
    handleMouseUp(event) {
        // console.warn('handleMouseUp', event.pageY);
        // Sometimes the mouse scrolls too fast to register the last hovered row.
        // If the mouseup position is higher than the datatable top, set lasthovered to 0
        if (this.dragging && this.draggingPos && this.draggingPos.bounding && event.pageY < this.draggingPos.bounding.top) {
            this.rowHoveredLast = 0;
        }
        // If a drag event ended NOT on a row, fire the onrowmouseup event with the last hovered row
        if (this.dragging && this.rowHoveredLast !== null) {
            this.onRowMouseEvent({ type: 'mouseup', rowIndex: this.rowHoveredLast, event: event });
            // this.onRowMouseUp(this.rowHoveredLast, event);
            // Unselect all text after drag to prevent weird selection issues
            if (document.getSelection) {
                document.getSelection().removeAllRanges();
            }
            else if (window.getSelection) {
                window.getSelection().removeAllRanges();
            }
        }
        this.dragging = false;
    }
    /**
     * On Global mouse move
     * @param {?} event
     * @return {?}
     */
    handleMouseMove(event) {
        const /** @type {?} */ draggingPos = Object.assign({ hasMinSize: false }, this.draggingPos);
        if (draggingPos &&
            draggingPos.bounding &&
            draggingPos.bounding.top &&
            draggingPos.bounding.bottom &&
            draggingPos.bounding.left &&
            draggingPos.bounding.right &&
            draggingPos.height &&
            draggingPos.top &&
            draggingPos.startY &&
            draggingPos.startX) {
            // Set to local reference so they can be changed
            let /** @type {?} */ pageY = event.pageY;
            let /** @type {?} */ pageX = event.pageX;
            // Set top boundary
            if (pageY < draggingPos.bounding.top) {
                pageY = draggingPos.bounding.top;
            }
            // Set bottom boundary
            if (pageY > draggingPos.bounding.bottom) {
                pageY = draggingPos.bounding.bottom;
            }
            // Set left boundary
            if (pageX < draggingPos.bounding.left) {
                pageX = draggingPos.bounding.left;
            }
            // Set right boundary
            if (pageX > draggingPos.bounding.right) {
                pageX = draggingPos.bounding.right;
            }
            // Determine if this is a right to left drag or a left to right
            if (pageX >= draggingPos.startX) {
                draggingPos.width = pageX - draggingPos.startX - 2;
                draggingPos.left = draggingPos.startX;
            }
            else {
                draggingPos.width = draggingPos.startX - pageX;
                draggingPos.left = pageX + 2;
            }
            // Only allow height and top changes if the drag is within the horizontal bounding box
            // This prevents the drag selection continuing to draw vertically even though
            // the mouse is off the datagrid which would give the user the impression
            // they are selecting rows even though their mouse is not on the grid and won't record the row mouse up event
            if (pageX < draggingPos.bounding.right && pageX > draggingPos.bounding.left) {
                // Determine if this is a top down or bottom up drag
                if (pageY >= draggingPos.startY) {
                    draggingPos.height = pageY - draggingPos.startY - 2;
                    draggingPos.top = draggingPos.startY;
                }
                else {
                    draggingPos.height = draggingPos.startY - pageY;
                    draggingPos.top = pageY + 2;
                }
            }
            // Make sure there's a minimum size so there isn't a FOUC
            if (draggingPos.width > 10 && draggingPos.height > 10) {
                draggingPos.hasMinSize = true;
            }
            // Make sure the top is lower than the bounding box, don't allow it to be dragged outside the box
            if (draggingPos.top < draggingPos.bounding.top) {
                draggingPos.top = draggingPos.bounding.top;
            }
            // Make sure the top is lower than the bounding box, don't allow it to be dragged outside the box
            if (draggingPos.height > draggingPos.bounding.top) {
                // draggingPos.top = draggingPos.bounding.top;
            }
            // Update DOM
            this.draggingPos = draggingPos;
        }
    }
    /**
     * Handle keyboard events
     * @param {?} event
     * @return {?}
     */
    handleKeyboardEvents(event) {
        // console.warn('handleKeyboardEvents 1');
        // Ignore keyboard repeat events
        if (event.repeat === false) {
            this.keyPressed = event.type === 'keydown' ? event.key : null;
            // If this is a keydown event, add it to the dictionary
            if (event.type === 'keydown') {
                this.keysPressed[event.key.toString().toLowerCase()] = true;
            }
            else if (event.type === 'keyup') {
                // If this is a key up event, remove from dictionary
                delete this.keysPressed[event.key.toString().toLowerCase()];
            }
            // If control + a is pressed, select all
            if (this.keysPressed['a'] && this.keysPressed['Control']) {
                this.selectRowsAll();
                event.preventDefault();
                event.stopPropagation();
            }
            // console.log('handleKeyboardEvents 1', event.type, this.keysPressed);
        }
    }
    /**
     * Select all rows
     * @return {?}
     */
    selectRowsAll() {
        if (this.rows.length) {
            this.rowsInternal.forEach(row => (row.$$selected = true));
            this.emitSelectedRows(this.rowsInternal);
        }
    }
    /**
     * When a mouse event has happend on a body row
     * @param {?} action
     * @return {?}
     */
    onRowMouseEvent(action) {
        if (action.type !== 'mouseenter') {
            // console.log('onRowMouseEvent', action);
        }
        const /** @type {?} */ row = this.rowsInternal.filter(row2 => row2.$$rowIndex === action.rowIndex)[0]; // this.rowsInternal[action.rowIndex];
        switch (action.type) {
            case 'click':
                this.selectRow(row, action.rowIndex, false);
                break;
            case 'contextmenu':
                this.selectRow(row, action.rowIndex, true);
                this.onRightClickMenu.emit(event); // Emit right click event up to the parent
                break;
            case 'mousedown':
                this.handleMouseDown(action.event);
                if (action.event.which === 1) {
                    // Only function when the left mouse button is clicked
                    this.rowClickDrag.rowIndex = action.rowIndex;
                }
                break;
            case 'mouseup':
                // Only function when the left mouse button is clicked
                if (action.event.which === 1 && action.rowIndex !== this.rowClickDrag.rowIndex) {
                    // Check if the drag was top to bottom or bottom to top for ROWS. Always start at the lowest index
                    const /** @type {?} */ rowStart = this.rowClickDrag.rowIndex <= action.rowIndex ? this.rowClickDrag.rowIndex : action.rowIndex;
                    const /** @type {?} */ rowEnd = this.rowClickDrag.rowIndex >= action.rowIndex ? this.rowClickDrag.rowIndex : action.rowIndex;
                    this.rowsInternal.forEach(rowNew => (rowNew.$$selected = false));
                    for (let /** @type {?} */ j = rowStart; j <= rowEnd; j++) {
                        this.rowsInternal[j].$$selected = true;
                    }
                    const /** @type {?} */ selectedRows = this.rowsInternal.filter(rowNew => rowNew.$$selected);
                    this.rowsSelectedCount = selectedRows.length;
                    this.emitSelectedRows(selectedRows);
                }
                break;
            case 'mouseenter':
                this.rowHoveredLast = action.rowIndex;
                break;
            case 'dblclick':
                this.rowHoveredLast = null;
                this.rowSelectedLast = null;
                break;
            default:
                console.warn('An unknown mouse event was passed to onRowMouseEvent');
        }
    }
    /**
     * Manage row selection. Includes single and multiple click by pressing control and shift
     * @param {?} row - The row object
     * @param {?} rowIndex - The index of the currently selected row
     * @param {?=} isRightClick
     * @param {?=} elementRef
     * @return {?}
     */
    selectRow(row, rowIndex, isRightClick, elementRef) {
        // console.warn('selectRow', this.keysPressed, row, rowIndex, isRightClick, elementRef);
        // Only allow row selection if set
        if (this.options.selectionType) {
            const /** @type {?} */ newRows = [...this.rowsInternal];
            // If control is pressed while clicking
            if (this.keysPressed['control'] && this.options.selectionType === 'multi') {
                row.$$selected = row.$$selected ? false : true;
            }
            else if (this.keysPressed['shift'] && this.options.selectionType === 'multi' && this.rowSelectedLast) {
                // If shift is pressed while clicking
                // Unset all selected flags
                newRows.forEach(rowNew => (rowNew.$$selected = false));
                // Figure out if the selection goes top to bottom or bottom to top
                const /** @type {?} */ startIndex = rowIndex > this.rowSelectedLast ? this.rowSelectedLast : rowIndex;
                const /** @type {?} */ endIndex = rowIndex < this.rowSelectedLast ? this.rowSelectedLast : rowIndex;
                // Loop through the lowest index and set all selected flags
                for (let /** @type {?} */ i = startIndex; i < endIndex + 1; i++) {
                    if (newRows[i]) {
                        newRows[i].$$selected = true;
                    }
                }
            }
            else if (!row.$$selected && !isRightClick) {
                // If just regular click
                // Disable all other selected flags
                newRows.forEach(rowNew => (rowNew.$$selected = false));
                row.$$selected = true;
                // this.rowsSelected = row;
            }
            else {
                // If this is a right click row, don't do anything special
                if (isRightClick) {
                    if (!row.$$selected) {
                        newRows.forEach(rowNew => (rowNew.$$selected = false));
                        row.$$selected = true;
                    }
                }
                else if (this.rowsSelectedCount > 1) {
                    // If multiple rows are already selected
                    // Reset all and set current row to selected
                    newRows.forEach(rowNew => (rowNew.$$selected = false));
                    row.$$selected = true;
                }
                else {
                    // Just untoggle this row
                    row.$$selected = false;
                }
            }
            const /** @type {?} */ selectedRows = newRows.filter(rowNew => rowNew.$$selected);
            this.rowSelectedLast = rowIndex;
            this.rowsSelectedCount = selectedRows.length;
            this.emitSelectedRows(selectedRows);
            this.onElementRef.emit(elementRef);
        }
    }
    /**
     * Create row css classes based on callback function in options
     * @param {?} row - Table row
     * @return {?}
     */
    getRowClasses(row) {
        if (!this.options.rowClass) {
            return null;
        }
        let /** @type {?} */ classes = '';
        const /** @type {?} */ results = this.options.rowClass(row);
        for (const /** @type {?} */ key in results) {
            if (results.hasOwnProperty(key)) {
                if (results[key]) {
                    classes += key + ' ';
                }
            }
        }
        if (classes !== '') {
            return classes;
        }
    }
    /**
     * Create a dictionary of row css classes based on inputs from options.rowClass
     * @return {?}
     */
    createRowClasses() {
        const /** @type {?} */ rowClasses = {};
        if (this.options.rowClass) {
            this.rows.forEach(row => {
                if (this.options.primaryKey && row[this.options.primaryKey] && this.options.rowClass) {
                    const /** @type {?} */ results = this.options.rowClass(row);
                    let /** @type {?} */ classes = '';
                    for (const /** @type {?} */ key in results) {
                        if (results[key]) {
                            classes += key + ' ';
                        }
                    }
                    rowClasses[row[this.options.primaryKey]] = classes.length ? classes : null;
                }
            });
        }
        this.rowClasses = rowClasses;
    }
    /**
     * Create inline css styles for rows
     * @return {?}
     */
    createRowStyles() {
        let /** @type {?} */ rowStyles = {};
        // Only create row styles if supplied by options
        if (!this.options.rowStyle) {
            this.rowStyles = rowStyles;
            return false;
        }
        const /** @type {?} */ primaryKey = this.options.primaryKey ? this.options.primaryKey : '';
        if (!this.options.primaryKey) {
            console.log('Please specify a primary key to use rowStyles');
        }
        // let primaryKey = this.options.primaryKey || '';
        /*
                    // If row height is set in options, set it in row styles
                    if (this.options.rowHeight) {
                  this.rows.forEach((row, index) => {
                    rowStyles[row[this.options.primaryKey]] =
        Object.assign(rowStyles[row[this.options.primaryKey]] || {},
        {'line-height': this.options.rowHeight + 'px'});
                  });
                }
                    */
        const /** @type {?} */ stylesWithModels = [];
        const /** @type {?} */ stylesNoModels = [];
        // Sort styles that have observable models and those that don't
        this.options.rowStyle.forEach(style => (style.model ? stylesWithModels.push(style.model) : stylesNoModels.push(style.rules)));
        // If no models
        if (stylesNoModels.length && primaryKey) {
            // Loop through all rows
            this.rows.forEach(row => {
                // Loop through all styles without rules
                stylesNoModels.forEach(rule => {
                    // Merge the newly created styles with what is already existing.
                    // This allows for multiple rulesets to assign styles without wiping out preexisting
                    rowStyles[row[primaryKey]] = Object.assign(rowStyles[row[primaryKey]], rule(row));
                });
            });
        }
        // If models/observables were supplied
        // This class method is NOT fired everytime observables update so it needs to be self contained
        // within the subscribe in order to update everything at once
        // Includes styles with no models in order to accomodate styles with mixed observables and non observables
        if (stylesWithModels.length) {
            // Create a single combine latest observable that will update when any of the inputs are updated
            const /** @type {?} */ subscription = combineLatest(stylesWithModels).subscribe((models) => {
                // Row styles needs to be complete refreshed everytime an observable changes
                rowStyles = {};
                /*
                                // If row height is set in options, set it in row styles
                                if (this.options.rowHeight) {
                                  this.rows.forEach((row, index) => {
                                    rowStyles[row[this.options.primaryKey]] =
                                Object.assign(rowStyles[row[this.options.primaryKey]] || {},
                                    {'line-height': this.options.rowHeight + 'px'});
                                  });
                                }
                                        */
                // If no models
                if (stylesNoModels.length && primaryKey) {
                    // Loop through all rows
                    this.rows.forEach(row => {
                        // Loop through all styles without rules
                        stylesNoModels.forEach(rule => {
                            // Merge the newly created styles with what is already existing.
                            //                This allows for multiple rulesets to assign styles without wiping out preexisting
                            rowStyles[row[primaryKey]] = Object.assign(rowStyles[row[primaryKey]] || {}, rule(row));
                        });
                    });
                }
                // Loop through all rays
                this.rows.forEach(row => {
                    // Now loop through each result of data returned from the observable
                    models.forEach((model, index) => {
                        if (this.options &&
                            this.options.rowStyle &&
                            this.options.rowStyle[index] &&
                            this.options.rowStyle[index].rules) {
                            // Merge the newly created styles with what is already existing.
                            //                This allows for multiple rulesets to assign styles without wiping out preexisting
                            if (this.options.rowStyle && this.options.rowStyle[index] && this.options.rowStyle[index].rules) {
                                rowStyles[row[primaryKey]] = Object.assign({}, (rowStyles[row[primaryKey]] || {}), (/** @type {?} */ (this.options.rowStyle[index])).rules(row, model));
                            }
                        }
                    });
                });
                // Update row styles
                this.rowStyles = rowStyles;
                // Tell DOM to updated after observable is done udpated
                this.ref.detectChanges();
            });
            this.subscriptions.push(/** @type {?} */ (subscription));
        }
        this.rowStyles = rowStyles;
    }
    /**
     * Emit changed columns up to the parent component
     * @return {?}
     */
    emitColumns() {
        // TODO: Mapping properties back up isn't seamless and needs work, commenting out for now
        // let remapColumns = this.dgSvc.mapPropertiesUp([...columns], this.options.columnMap);
        // Remove templates and emit new column references up. Templates have a circulate reference which blows up json usage
        const /** @type {?} */ columnsNew = [...this.columnsPinnedLeft, ...this.columnsInternal];
        const /** @type {?} */ columnsEmitted = columnsNew.map(column => {
            const /** @type {?} */ columnNew = Object.assign({}, column);
            columnNew.locked = columnNew.pinnedLeft ? true : false;
            delete columnNew.templateCell;
            delete columnNew.templateHeader;
            return columnNew;
        });
        // Emit data back up
        this.onColumnsUpdated.emit(columnsEmitted);
    }
    /**
     * Emit state changes up to the parent component
     * @param {?} state
     * @return {?}
     */
    emitState(state) {
        // User columns has a circular reference somewhere so create new instance and remove that property before emitting up
        const /** @type {?} */ stateNew = Object.assign({}, state);
        delete (/** @type {?} */ (stateNew)).usersColumns;
        // Create a new memory reference for the state and then remap all properties up into the layout
        const /** @type {?} */ remapProps = JSON.parse(JSON.stringify(stateNew));
        remapProps.groups =
            remapProps.groups && remapProps.groups.length
                ? this.dgSvc.mapPropertiesUp(remapProps.groups, this.options.controlsMap)
                : [];
        remapProps.sorts =
            remapProps.sorts && remapProps.sorts.length
                ? this.dgSvc.mapPropertiesUp(remapProps.sorts, this.options.controlsMap)
                : [];
        remapProps.filters =
            remapProps.filters && remapProps.filters.length
                ? this.dgSvc.mapPropertiesUp(remapProps.filters, this.options.controlsMap)
                : [];
        this.onStateChange.emit(remapProps);
    }
    /**
     * When a row was edited
     * @param {?} event
     * @return {?}
     */
    rowUpdated(event) {
        this.onRowUpdated.emit(event);
    }
    /**
     * Pass selected rows up to the parent component after cleaning up any DT2 properties
     * @param {?} rows
     * @return {?}
     */
    emitSelectedRows(rows) {
        /** //Removed for now, we need the instance passed up so the parent can manipulate the rows
                  // Loop through all the selected rows and remove any $$ properties
                  let newRows = rows.map(row => {
                    // Return new instance
                    let newRow = Object.assign({}, row);
                    // Loop through all keys in object
                    for (let key in newRow) {
                      // Remove any keys with $$ present
                      if (key.indexOf('$$') != -1){
                        delete newRow[key];
                      }
                    }
                    return newRow;
                  });
                  */
        this.onRowsSelected.emit([...rows]);
    }
    /**
     * Perform an action on the main datatable that was requested by lower component
     * @param {?} action
     * @return {?}
     */
    onUpdateDatatable(action) {
        // Update datatable
        if (action === 'update') {
            this.viewCreate();
        }
        else if (action === 'reset') {
            // Reset datatable
            this.reset();
        }
    }
    /**
     * Emit a custom link event response up to the parent component
     * @param {?} data
     * @return {?}
     */
    customLinkEvent(data) {
        this.onCustomLinkEvent.emit(data);
    }
    /**
     * Reset all datatable controls, filters sorts groups etc
     * @param {?=} resetType
     * @return {?}
     */
    reset(resetType) {
        this.ref.detach();
        // Reset State
        if (resetType && resetType === 'groups') {
            this.state.groups = [];
        }
        else if (resetType && resetType === 'sorts') {
            this.state.sorts = [];
        }
        else if (resetType && resetType === 'filters') {
            this.state.filters = [];
        }
        else {
            this.state.groups = [];
            this.state.filters = [];
            this.state.sorts = [];
        }
        this.state.info = {};
        // Reset Columns
        this.columnsInternal = this.columnsInternal.map(column => {
            column.pinnedLeft = false;
            column.locked = false;
            return Object.assign({}, column);
        });
        if (this.filterGlobal) {
            this.filterGlobal.term = '';
        }
        this.emitColumns();
        this.onStateUpdated({ action: Actions.reset, data: null });
        this.ref.reattach();
    }
    /**
     * On window resize
     * @return {?}
     */
    onWindowResize() {
        if (this.columnsInternal && this.columnsInternal.length && this.rowsInternal && this.rowsInternal.length) {
            this.viewCreate();
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        // Unsub from all subscriptions
        this.subscriptions.forEach(sub => sub.unsubscribe());
    }
}
DataGridComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid',
                template: `
<div id="datagrid" class="datagrid" #dataGrid
	 [ngClass]="{
    'selectable' : options.selectionType,
    'truncatable' : options.columnsTruncate,
    'fullscreen' : options.fullScreen }">

	<div *ngIf="!appReady || !domReady" class="container-fluid pt-3" style="font-size: 1rem;"><i class="fa fa-spinner fa-spin"></i> Loading View...</div>

	<div class="datagrid-container" *ngIf="appReady && domReady">

		<!--TODO: Add readonly div -->
		<!-- Header -->
		<datagrid-header [columns]="columnsExternal"
						 [columnsPinnedLeft]="columnsPinnedLeft"
						 [state]="state"
						 [status]="status"
						 [options]="options"
						 [scrollProps]="scrollProps"
						 [gridProps]="gridProps"
						 [filterTerms]="filterTerms"
             [templates]="columnTemplates"
						 (onStateUpdated)="onStateUpdated($event)"
						 (onColumnsUpdated)="columnsUpdated($event)"
						 (onCustomLinkEvent)="customLinkEvent($event)">
		</datagrid-header>

		<!-- Info bar-->
		<datagrid-info *ngIf="options.showInfo && (state?.sorts.length || state?.groups.length || state?.filters.length)"
					   [state]="state"
					   [options]="options"
					   [columnsMapped]="columnsMapped"
					   (onReset)="reset($event)">
		</datagrid-info>

		<!-- Body -->
		<datagrid-body #dataGridBody
					   [columns]="columnsExternal"
					   [columnsPinnedLeft]="columnsPinnedLeft"
					   [rows]="rowsExternal"
					   [rowStyles]="rowStyles"
					   [state]="state"
					   [status]="status"
					   [options]="options"
					   [gridProps]="gridProps"
             [templates]="columnTemplates"
					   (onRowMouseEvent)="onRowMouseEvent($event)"
					   (onRowUpdated)="rowUpdated($event)"
					   (onGroupToggled)="groupToggled($event)"
					   (onScrollEvent)="onScroll($event)">
		</datagrid-body>
	</div>

	<!-- Drag select box -->
	<div id="drag-select"
		 *ngIf="options.showDragBox && dragging && options.selectionType && draggingPos.hasMinSize"
		 [ngStyle]="{ 'top':draggingPos.top + 'px', 'left': draggingPos.left + 'px', width: draggingPos.width + 'px', height: draggingPos.height + 'px' }">
	</div>

</div>
`,
                styles: [`#datagrid{font-size:12px;border-top:1px solid #ccc;position:relative}#datagrid.maxHeight{overflow-y:scroll;overflow-x:hidden}#datagrid.fullscreen{position:absolute;top:0;bottom:0;left:0;right:0}#datagrid.scrollH{overflow-x:scroll}#datagrid.selectable{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}#datagrid.selectable .datagrid-body-row:hover:before{content:'';display:block;position:absolute;top:-1px;height:2px;z-index:999;left:0;right:0;background:#000}#datagrid.selectable .datagrid-body-row:hover:after{content:'';display:block;position:absolute;bottom:-2px;height:2px;z-index:999;left:0;right:0;background:#000}#datagrid.selectable .datagrid-body-row.selected,#datagrid.selectable .datagrid-body-row.selected .datagrid-body-pinned{background-color:#ececec}#datagrid.selectable .datagrid-body-row.selected:before{content:'';display:block;position:absolute;top:-1px;height:2px;z-index:999;left:0;right:0;background:red}#datagrid.selectable .datagrid-body-row.selected:after{content:'';display:block;position:absolute;bottom:-2px;height:2px;z-index:999;left:0;right:0;background:red}#datagrid.selectable .datagrid-body-row.selected+.datagrid-body-row.selected:before{display:none!important}#datagrid .sticky-left{position:-webkit-sticky;position:sticky;left:0}#datagrid #drag-select{width:20px;height:20px;background:red;position:fixed;opacity:.3;z-index:1000}#datagrid .datagrid-container{margin:0;background-color:#fff;overflow:hidden;min-height:200px}#datagrid .datagrid-container .datagrid-header-container{height:24px;position:relative;z-index:4;background-image:-webkit-gradient(linear,left bottom,left top,from(#dadada),to(#f1f1f1));background-image:linear-gradient(to top,#dadada 0,#f1f1f1 100%)}#datagrid .datagrid-container .datagrid-header .datagrid-header-row{background-color:#f1f1f1;display:-webkit-box;display:-ms-flexbox;display:flex;-webkit-box-orient:horizontal;-webkit-box-direction:normal;-ms-flex-flow:row;flex-flow:row;-webkit-backface-visibility:hidden;backface-visibility:hidden}#datagrid .datagrid-container .datagrid-header .datagrid-header-pinned{left:0;z-index:7!important;background-color:#f1f1f1;display:-webkit-box;display:-ms-flexbox;display:flex;-webkit-box-orient:horizontal;-webkit-box-direction:normal;-ms-flex-flow:row;flex-flow:row}#datagrid .datagrid-container .datagrid-header .datagrid-header-main{display:-webkit-box;display:-ms-flexbox;display:flex;-webkit-box-orient:horizontal;-webkit-box-direction:normal;-ms-flex-flow:row;flex-flow:row;position:absolute}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell{line-height:20px;font-weight:700;display:inline-block;cursor:default;background-color:#f1f1f1;background-image:-webkit-gradient(linear,left bottom,left top,from(#dadada),to(#f1f1f1));background-image:linear-gradient(to top,#dadada 0,#f1f1f1 100%);white-space:nowrap;padding:0;outline:#ccc solid 1px;top:0;position:absolute;z-index:6;border:0}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell.pinned{z-index:7!important;background-color:#f1f1f1}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell.reorderable .datagrid-header-cell-label{cursor:grab;cursor:-webkit-grab}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell.column-last .dropdown-submenu-flyout{left:auto!important;right:100%!important}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell>div{padding-right:8px}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell .resize-handle{display:block;position:absolute;background:#f6ac38;top:0;bottom:0;right:-2px;width:5px;z-index:8;cursor:ew-resize}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell .datagrid-header-cell-label{padding:2px 8px;overflow:hidden;display:block}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell a.active{color:red!important}#datagrid .datagrid-container .datagrid-header .datagrid-header-cell .header-controls{display:inline-block!important;position:absolute;right:0;padding-right:5px;padding-left:5px;background-color:#f1f1f1;background-image:-webkit-gradient(linear,left bottom,left top,from(#dadada),to(#f1f1f1));background-image:linear-gradient(to top,#dadada 0,#f1f1f1 100%);top:50%;-webkit-transform:translateY(-50%);transform:translateY(-50%)}#datagrid .datagrid-container .datagrid-header .dropdown .active{color:red;background:0 0}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-item{padding:3px .75rem;font-size:.8rem}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-item:hover{background-color:#f7f7f9}#datagrid .datagrid-container .datagrid-header .dropdown .disabled:hover{background-color:transparent}#datagrid .datagrid-container .datagrid-header .dropdown>.dropdown-menu{top:20px!important;min-width:200px;font-size:.8rem;padding:.5rem 0}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu{position:relative;display:block;font-size:.8rem}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .dropdown-submenu-button:after{font-family:FontAwesome;content:'\\f0da';position:absolute;top:50%;-webkit-transform:translateY(-50%);transform:translateY(-50%);right:10px}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .dropdown-submenu-flyout{left:100%;top:-25px!important;display:none;padding-top:0}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .dropdown-submenu-flyout .btn-group{width:100%}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .dropdown-submenu-flyout .btn{width:50%}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu.dropdown-submenu-left .dropdown-submenu-flyout{right:100%;left:auto}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu:hover .dropdown-submenu-flyout{display:block;padding:.5rem 0}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .terms-list{max-height:300px;overflow-y:auto;padding-bottom:.75rem}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .terms-list .term{font-weight:400;font-size:.9rem;padding:0 2rem}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .terms-list label.term:hover{background-color:#f7f7f9}#datagrid .datagrid-container .datagrid-header .dropdown .dropdown-submenu .box{padding:0 1rem}#datagrid .datagrid-container .datagrid-body-container{overflow:scroll;position:relative;z-index:3}#datagrid .datagrid-container .datagrid-body-container.widthFixed{overflow-x:hidden}#datagrid .datagrid-container .datagrid-info-row{line-height:23px;padding-left:.5rem;border-bottom:1px solid #ccc;background-color:#ededed;-webkit-box-shadow:5px 1px 5px rgba(0,0,0,.3);box-shadow:5px 1px 5px rgba(0,0,0,.3);overflow:initial}#datagrid .datagrid-container .datagrid-body .datagrid-body-row{display:-webkit-box;display:-ms-flexbox;display:flex;-webkit-box-orient:horizontal;-webkit-box-direction:normal;-ms-flex-flow:row;flex-flow:row;-webkit-backface-visibility:hidden;backface-visibility:hidden;position:absolute;border-bottom:1px solid #b3b3b3;height:24px}#datagrid .datagrid-container .datagrid-body .datagrid-body-pinned{left:0;background-color:#fff;z-index:4;display:inline-block;border-right:1px solid #ccc}#datagrid .datagrid-container .datagrid-body .datagrid-body-main{min-height:23.5px}#datagrid .datagrid-container .datagrid-body .datagrid-row{position:relative;display:-webkit-box;display:-ms-flexbox;display:flex;-webkit-box-orient:horizontal;-webkit-box-direction:normal;-ms-flex-flow:row;flex-flow:row;min-height:23px}#datagrid .datagrid-container .datagrid-body .datagrid-row.selected-state .datagrid-cell{background:#f0dada}#datagrid .datagrid-container .datagrid-body .datagrid-row .datagrid-cell{position:absolute;min-height:23px;border:none}#datagrid .datagrid-container .datagrid-body .datagrid-row .datagrid-cell-data{line-height:23px;position:relative;z-index:3;display:inline-block;white-space:nowrap;text-overflow:ellipsis;min-height:23px;overflow:hidden}#datagrid .datagrid-container .datagrid-body .datagrid-row .datagrid-cell-data .datagrid-cell-data.canEdit:hover{outline:#e0e0e0 dashed 2px;outline-offset:-2px}#datagrid .datagrid-container .datagrid-body .datagrid-row .datagrid-cell-data .datagrid-cell-data-content{position:absolute;padding:0 5px;left:0;right:0;bottom:0;top:0}#datagrid .datagrid-container .datagrid-body .datagrid-row .datagrid-cell-data.pinned{-webkit-box-shadow:2px 5px 5px rgba(0,0,0,.3);box-shadow:2px 5px 5px rgba(0,0,0,.3);background-color:#fff;z-index:4}#datagrid .datagrid-container .datagrid-body .datagrid-row .datagrid-cell-data .cell{height:22px;padding:3px 8px}#datagrid .datagrid-container .datagrid-body .datagrid-row .datagrid-cell-data .cell-tooltip{position:absolute;left:0;right:0;top:0;bottom:0}#datagrid .datagrid-container .datagrid-body .datagrid-truncated{width:5px;height:5px;background:red;-webkit-transform:rotate(45deg);transform:rotate(45deg);position:absolute;top:-2px;right:-2px;overflow:hidden}#datagrid .datagrid-container .datagrid-body .datagrid-group-header{min-height:24px}#datagrid .datagrid-container .datagrid-body .datagrid-group-header .datagrid-group-header-cell{padding:0;z-index:4;left:0}#datagrid .datagrid-container .datagrid-body .datagrid-group-header a{background-image:-webkit-gradient(linear,left bottom,left top,from(#fcfeff),to(#cbdce6));background-image:linear-gradient(to top,#fcfeff 0,#cbdce6 100%);padding:.15rem .15rem .15rem .5rem;display:block;border-bottom:1px solid #c2d1da}#datagrid .datagrid-container .datagrid-body .datagrid-group-header a>span{left:16px}#datagrid .datagrid-container .datagrid-body .datagrid-group-header a>span:before{font-family:FontAwesome;content:'\\f0da';-webkit-transition:.1s ease-out;transition:.1s ease-out;display:inline-block;margin-right:5px}#datagrid .datagrid-container .datagrid-body .datagrid-group-header a.active span:before{-webkit-transform:rotate(90deg);transform:rotate(90deg);-webkit-transition:.1s ease-out;transition:.1s ease-out}#datagrid .fa-stack-groupsort{margin-right:-5px}#datagrid .fa-stack-groupsort .arrow{margin-right:-3px;font-size:104%}#datagrid .fa-stack-groupsort .group{position:relative;top:-1px}#datagrid .anim-fade-in{opacity:0;-webkit-animation:.3s ease-in forwards fade;animation:.3s ease-in forwards fade;position:relative}@-webkit-keyframes fade{from{opacity:0}to{opacity:1}}@keyframes fade{from{opacity:0}to{opacity:1}}`],
                encapsulation: ViewEncapsulation.None,
                changeDetection: ChangeDetectionStrategy.OnPush,
                host: {
                    '(document:keydown)': 'handleKeyboardEvents($event)',
                    '(document:keyup)': 'handleKeyboardEvents($event)',
                    // '(document:mousedown )': 'handleMouseDown($event)',
                    // '(document:mouseup)': 'handleMouseUp($event)',
                    // '(document:mousemove)': 'handleMouseMove($event)',
                    '(window:resize)': 'onWindowResizeThrottled($event)',
                },
            },] },
];
/** @nocollapse */
DataGridComponent.ctorParameters = () => [
    { type: DataGridService, },
    { type: ChangeDetectorRef, },
];
DataGridComponent.propDecorators = {
    "dataGrid": [{ type: ViewChild, args: ['dataGrid',] },],
    "dataGridBody": [{ type: ViewChild, args: ['dataGridBody',] },],
    "columns": [{ type: Input },],
    "rows": [{ type: Input },],
    "state": [{ type: Input },],
    "columnTemplates": [{ type: ContentChildren, args: [DataTableColumnDirective,] },],
    "options": [{ type: Input },],
    "filterGlobal": [{ type: Input },],
    "onColumnsUpdated": [{ type: Output },],
    "onRowsSelected": [{ type: Output },],
    "onStateChange": [{ type: Output },],
    "onRightClickMenu": [{ type: Output },],
    "action": [{ type: Output },],
    "onCustomLinkEvent": [{ type: Output },],
    "onElementRef": [{ type: Output },],
    "onRowUpdated": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class HeaderComponent {
    /**
     * During a resize, disable some stuff
     */
    constructor() {
        this.onColumnsUpdated = new EventEmitter();
        this.onStateUpdated = new EventEmitter();
        this.onCustomLinkEvent = new EventEmitter();
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * Pass state changes up from controls component
     * @param {?} event
     * @return {?}
     */
    stateUpdated(event) {
        this.onStateUpdated.emit(event);
    }
    /**
     * Pass column changes up to the main datagrid
     * @param {?} event
     * @return {?}
     */
    columnsUpdated(event) {
        this.onColumnsUpdated.emit(event);
    }
    /**
     * Emit a custom link event response up to the parent component
     * @param {?} data
     * @return {?}
     */
    customLinkEvent(data) {
        this.onCustomLinkEvent.emit(data);
    }
}
HeaderComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid-header',
                template: `<div class="datagrid-header-container" [ngStyle]="{'width': gridProps.widthViewPort + 'px'}">
    <div class="datagrid-header" [ngStyle]="{'width' : gridProps.widthTotal + 'px'}">

        <!-- Header Row -->
        <div class="datagrid-header-row">

            <div class="datagrid-header-pinned sticky-left" *ngIf="columnsPinnedLeft?.length" [ngStyle]="{'width' : gridProps.widthPinned + 'px' }">
                <datagrid-header-row [columns]="columnsPinnedLeft"
                                     [columnType]="'pinnedLeft'"
                                     [state]="state"
                                     [status]="status"
                                     [options]="options"
                                     [scrollProps]="scrollProps"
                                     [filterTerms]="filterTerms"
                                     [templates]="templates"
                                     (onColumnsUpdated)="columnsUpdated($event)"
                                     (onStateUpdated)="stateUpdated($event)"
                                     (onCustomLinkEvent)="customLinkEvent($event)">
                </datagrid-header-row>
            </div>
            
            <div class="datagrid-header-main" [ngStyle]="{ 'left' : (-scrollProps.scrollLeft + gridProps.widthPinned) + 'px' }">
                <datagrid-header-row [columns]="columns"
                                     [columnType]="'main'"
                                     [state]="state"
                                     [status]="status"
                                     [options]="options"
                                     [scrollProps]="scrollProps"
                                     [filterTerms]="filterTerms"
                                     [templates]="templates"
                                     (onColumnsUpdated)="columnsUpdated($event)"
                                     (onStateUpdated)="stateUpdated($event)"
                                     (onCustomLinkEvent)="customLinkEvent($event)">
                </datagrid-header-row>
            </div>

        </div>
    </div>
</div>
`,
                styles: [],
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/**
    * On a successfull drag reorder of the column headers
    public onReorderSuccess() {
        // If columns are being dragged before a pinned column, set that column to pinned
        let isPinned = false;
        for (let i = this.columns.length - 1; i >= 0; i--) {
            let column = this.columns[i];
            if (column.pinnedLeft) {
                isPinned = true;
            }
            column.locked = isPinned;
            column.pinnedLeft = isPinned;
        }
        this.onColumnsUpdated.emit(this.columns);
    }
    */
/** @nocollapse */
HeaderComponent.ctorParameters = () => [];
HeaderComponent.propDecorators = {
    "columns": [{ type: Input },],
    "columnsPinnedLeft": [{ type: Input },],
    "state": [{ type: Input },],
    "status": [{ type: Input },],
    "options": [{ type: Input },],
    "gridProps": [{ type: Input },],
    "scrollProps": [{ type: Input },],
    "filterTerms": [{ type: Input },],
    "templates": [{ type: Input },],
    "onColumnsUpdated": [{ type: Output },],
    "onStateUpdated": [{ type: Output },],
    "onCustomLinkEvent": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class FiltersComponent {
    constructor() {
        this.filterTerms = {};
        this.onFiltersUpdated = new EventEmitter();
        this.filterTerm = new BehaviorSubject(null);
        this.subs = [];
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        // Create an subscription to debounce the user input
        this.filterTerm.pipe(debounceTime(300)).subscribe(filter => {
            if (filter) {
                this.onFiltersUpdated.emit(filter);
            }
        });
        // If custom column data has been supplied
        if (this.options.columnData) {
            const /** @type {?} */ sub = this.options.columnData[this.column.prop].model.subscribe(model => {
                if (this.options.columnData) {
                    // If a modelSrc has been supplied, fetch data from that. If not get it straight from the model
                    this.model = this.options.columnData[this.column.prop].modelSrc
                        ? model[this.options.columnData[this.column.prop].modelSrc]
                        : model;
                    // Simplify some of the properties for easier use in the dom
                    this.modelLabel = this.options.columnData[this.column.prop].label || '';
                    this.modelValue = this.options.columnData[this.column.prop].value || '';
                    this.modelClasses = this.options.columnData[this.column.prop].classes || '';
                    this.modelStyles = this.options.columnData[this.column.prop].styles || '';
                }
            });
            this.subs.push(sub);
        }
    }
    /**
     * When the open form filter is modified
     * @param {?} columnProp
     * @param {?} operator
     * @param {?} newTerm
     * @param {?} oldTerm
     * @return {?}
     */
    modifyFilter(columnProp, operator, newTerm, oldTerm) {
        // console.warn('modifyFilter',newTerm, oldTerm);
        const /** @type {?} */ filterObj = {
            filterAction: '',
            filter: { prop: columnProp, operator: operator, value: newTerm },
        };
        // Adding new filter
        if ((oldTerm === '' || !oldTerm) && newTerm !== '') {
            filterObj.filterAction = 'add';
        }
        else if (newTerm === '' || newTerm === false || !newTerm || newTerm === oldTerm) {
            // Removing existing filter
            filterObj.filterAction = 'remove';
        }
        else if (newTerm !== oldTerm) {
            // Changing existing filter
            filterObj.filterAction = 'change';
        }
        // console.warn('modifyFilter 2', JSON.parse(JSON.stringify(filterObj)));
        this.filterTerm.next(filterObj);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        if (this.subs.length) {
            this.subs.forEach(sub => sub.unsubscribe());
        }
    }
}
FiltersComponent.decorators = [
    { type: Component, args: [{
                selector: '<filters>',
                template: `<!--
<div class="btn-group" role="group">
        <button type="button" class="btn btn-interact-outline"><i aria-hidden="true" class="fa fa-filter"></i> Filter</button>
        <button type="button" class="btn btn-interact-outline"><i aria-hidden="true" class="fa fa-remove"></i> Clear</button>
</div>
-->

 <!-- Filtering -->
 <div *ngIf="column.canSearch">
    <label class="icon filter ml-2 mr-2 mt-2" *ngIf="status.filters[column.prop]">
        <input class="form-control form-control-sm" placeholder="Filter" name="filterSearch" [ngModel]="status.filters[column.prop].contains" (ngModelChange)="modifyFilter(column.prop, 'contains', $event, status.filters[column.prop].contains)"  [attr.data-qa]="'controls-filter-search'"/>
        <a class="clear" (click)="modifyFilter(column.prop, 'contains', '', status.filters[column.prop].contains)" *ngIf="status.filters[column.prop].contains != ''">
            <i class="fa fa-remove"></i></a>
    </label>
</div>

<!-- Strings -->
<ng-container *ngIf="column.columnType == 'string'">
    <div class="terms-list" [attr.data-qa]="'controls-filter-terms'">
        <ng-container *ngIf="options?.columnData && options?.columnData[column.prop];else filterList">
            <label class="term form-check-label field-{{column.prop + '-' + term[modelValue]}} {{term[modelClasses]}}" [ngStyle]="term[modelStyles]" *ngFor="let term of model">
                <input type="checkbox" class="form-check-input" [value]="term[modelValue]"
                [ngModel]="status.filters[column.prop]?.eq[term[modelValue]]" (ngModelChange)="modifyFilter(column.prop, 'eq', term[modelValue], status.filters[column.prop].eq[term[modelValue]])">
                <span>{{term[modelLabel]}}</span>
            </label>
        </ng-container>
        <ng-template #filterList>
            <ng-container *ngIf="filterTerms[this.column.prop]?.length">
                <label class="term form-check-label" *ngFor="let term of filterTerms[this.column.prop]">
                    <input type="checkbox" class="form-check-input" [value]="term" [attr.data-qa]="'controls-filter-term' + term"
                           [ngModel]="status.filters[column.prop]?.eq[term]" (ngModelChange)="modifyFilter(column.prop, 'eq', term, status.filters[column.prop].eq[term])">
                    <span>{{term}}</span>
                </label>
            </ng-container>
            <span *ngIf="!filterTerms[this.column.prop]?.length" class="term">No filter terms found</span>
        </ng-template>
        
    </div>
</ng-container>

<!-- Numbers and Money 
<ng-container *ngIf="column.columnType == 'int' || column.columnType == 'money'">
	<h6 class="dropdown-header">Show items with values that:</h6>
	<div class="box">
		<div class="form-group">
			<select class="form-control form-control-sm mb-1">
				<option>Is Equal To</option>
				<option>Is Not Equal To</option>
				<option>Is Greater Than Or Equal To</option>
				<option>Is Greater Than</option>
				<option>Is Less Than Or Equal To</option>
				<option>Is Less Than</option>
				<option>Is Null</option>
				<option>Is Not Null</option>
			</select>
			<input class="form-control form-control-sm" type="number" />
		</div>
		<div class="form-group">
			<select class="form-control form-control-sm">
				<option>And</option>
				<option>Or</option>
			</select>
		</div>
		<div class="form-group">
			<select class="form-control form-control-sm mb-1">
				<option>Is Equal To</option>
				<option>Is Not Equal To</option>
				<option>Is Greater Than Or Equal To</option>
				<option>Is Greater Than</option>
				<option>Is Less Than Or Equal To</option>
				<option>Is Less Than</option>
				<option>Is Null</option>
				<option>Is Not Null</option>
			</select>
			<input class="form-control form-control-sm" type="number" />
		</div>
	</div>
</ng-container>
-->

<!-- Dates
<ng-container *ngIf="column.columnType == 'date'">
    <h6 class="dropdown-header">Show items with values that:</h6>
    <div class="box">
        <div class="form-group">
            <select class="form-control form-control-sm mb-1" [(ngModel)]="model" >
                <option>Is Equal To</option>
                <option>Is Not Equal To</option>
                <option>Is Greater Than Or Equal To</option>
                <option>Is Greater Than</option>
                <option>Is Less Than Or Equal To</option>
                <option>Is Less Than</option>
                <option>Is Null</option>
                <option>Is Not Null</option>
            </select>
            <form class="form-inline">
                <div class="form-group w-100">
                    <div class="input-group">
                        <input class="form-control" placeholder="yyyy-mm-dd"
                               name="dp" [(ngModel)]="model" ngbDatepicker #d="ngbDatepicker">
                        <button class="input-group-addon" (click)="d.toggle()" type="button">
                            <i class="fa fa-calendar"></i>
                        </button>
                    </div>
                </div>
            </form>
        </div>
        <div class="form-group">
            <select class="form-control form-control-sm">
                <option>And</option>
                <option>Or</option>
            </select>
        </div>
        <div class="form-group">
            <select class="form-control form-control-sm mb-1">
                <option>Is Equal To</option>
                <option>Is Not Equal To</option>
                <option>Is Greater Than Or Equal To</option>
                <option>Is Greater Than</option>
                <option>Is Less Than Or Equal To</option>
                <option>Is Less Than</option>
                <option>Is Null</option>
                <option>Is Not Null</option>
            </select>
            <form class="form-inline">
                <div class="form-group w-100">
                    <div class="input-group">
                        <input class="form-control" placeholder="yyyy-mm-dd"
                               name="dp" [(ngModel)]="model2" ngbDatepicker #d2="ngbDatepicker">
                        <button class="input-group-addon" (click)="d.toggle()" type="button">
                            <i class="fa fa-calendar"></i>
                        </button>
                    </div>
                </div>
            </form>
        </div>
        
    </div>
</ng-container>
-->
`,
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
FiltersComponent.ctorParameters = () => [];
FiltersComponent.propDecorators = {
    "column": [{ type: Input },],
    "options": [{ type: Input },],
    "state": [{ type: Input },],
    "status": [{ type: Input },],
    "filterTerms": [{ type: Input },],
    "onFiltersUpdated": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ControlsComponent {
    constructor() {
        this.onStateUpdated = new EventEmitter();
        this.onCustomLinkEvent = new EventEmitter();
        this.subs = [];
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * When the modify sort button is clicked
     * @param {?} action
     * @param {?} prop
     * @param {?=} direction
     * @return {?}
     */
    modifySorts(action, prop, direction) {
        if (!direction) {
            direction = 'asc';
        }
        else if (direction === 'asc') {
            direction = 'desc';
        }
        else if (direction === 'desc') {
            direction = null;
        }
        this.modifyState(action, { dir: direction, prop: prop });
    }
    /**
     * Column pinning
     * @param {?} action
     * @param {?} column
     * @param {?=} index
     * @return {?}
     */
    modifyPinned(action, column, index) {
        this.modifyState(action, { prop: column.prop, index: index, isPinned: column.pinnedLeft });
    }
    /**
     * Modify filter state
     * @param {?} data
     * @return {?}
     */
    modifyFilters(data) {
        this.modifyState(Actions.filter, data);
    }
    /**
     * Clear all filters for this column
     * @param {?} columnProp
     * @return {?}
     */
    clearFilters(columnProp) {
        this.modifyState(Actions.filter, { filterAction: 'clear', filter: { prop: columnProp } });
    }
    /**
     * When the modify state button is clicked
     * @param {?} action
     * @param {?} data
     * @return {?}
     */
    modifyState(action, data) {
        this.onStateUpdated.emit({ action: action, data: data });
    }
    /**
     * When a custom link is clicked, emit that value up to the parent
     * @param {?} link
     * @param {?} column
     * @return {?}
     */
    customLinkAction(link, column) {
        this.onCustomLinkEvent.emit({ link: link, column: column });
    }
    /**
     * Delete a column
     * @param {?} columnIndex
     * @return {?}
     */
    deleteColumn(columnIndex) {
        this.modifyState(Actions.column, { action: 'delete', columnIndex: columnIndex });
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        if (this.subs.length) {
            this.subs.forEach(sub => sub.unsubscribe());
        }
    }
}
ControlsComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid-controls',
                template: `<div ngbDropdown id="headerComponentDropdown" class="header-controls d-inline-block dropdown" #myDrop="ngbDropdown" placement="{{columnIndex < 1 ? 'bottom-left' : 'bottom-right'}}">
    <a ngbDropdownToggle *ngIf="status?.groups || status?.sorts || status?.filters" [attr.data-qa]="'controls-dropdown'"
       [ngClass]="{'active' : status.groups[column.prop] || status.sorts[column.prop] || status.filters[column.prop]?.hasFilters || column.pinnedLeft }">
        <i class="fa fa-bars" aria-hidden="true"></i>
    </a>
    <div ngbDropdownMenu aria-labelledby="dropdownBasic1">

        <!-- Sorting -->
        <button class="dropdown-item" *ngIf="column.canSort && status?.sorts" (click)="modifySorts('sort', column.prop, status.sorts[column.prop])" [attr.data-qa]="'controls-sort'"
                [ngClass]="{'active' : status.sorts[column.prop] }">
            <ng-container *ngIf="status.sorts[column.prop] == 'asc'">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa fa-fw fa-sort-amount-asc" aria-hidden="true"></i>
                </span> Sorted Ascending
            </ng-container>
            <ng-container *ngIf="status.sorts[column.prop] == 'desc'">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa fa-fw fa-sort-amount-asc fa-flip-vertical" aria-hidden="true"></i>
                </span> Sorted Descending
            </ng-container>
            <ng-container *ngIf="!status.sorts[column.prop]">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa fa-fw fa-sort-amount-desc" aria-hidden="true"></i>
                </span> Sort By {{column.label}}
            </ng-container>
        </button>

        <!-- Grouping -->
        <button class="dropdown-item" *ngIf="column.canGroup && status?.groups" (click)="modifySorts('group', column.prop, status.groups[column.prop])" [attr.data-qa]="'controls-group'"
                [ngClass]="{'active' : status.groups[column.prop] }">
            <ng-container *ngIf="status.groups[column.prop] == 'asc'">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa arrow fa-long-arrow-down" aria-hidden="true"></i>
                    <i class="fa fa-object-group" aria-hidden="true"></i>
                </span> Grouped Ascending
            </ng-container>
            <ng-container *ngIf="status.groups[column.prop] == 'desc'">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa arrow fa-long-arrow-up" aria-hidden="true"></i>
                    <i class="fa fa-object-group" aria-hidden="true"></i>
                </span> Grouped Descending
            </ng-container>
            <ng-container *ngIf="!status.groups[column.prop]">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa arrow fa-long-arrow-down" aria-hidden="true"></i>
                    <i class="fa fa-object-group" aria-hidden="true"></i>
                </span> Group By {{column.label}}
            </ng-container>
        </button>

        <!-- Pinning -->
        <button class="dropdown-item" *ngIf="column.canPinLeft" (click)="myDrop.close();modifyPinned('pinLeft', column, columnIndex)" [attr.data-qa]="'controls-pinLeft'"
                [ngClass]="{'active' : column.pinnedLeft }">
            <ng-container *ngIf="column.pinnedLeft">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa fa-caret-square-o-left" aria-hidden="true"></i>
                </span> Pinned
            </ng-container>
            <ng-container *ngIf="!column.pinnedLeft">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa fa-caret-square-o-left" aria-hidden="true"></i>
                </span> Pin {{column.label}}
            </ng-container>
        </button>

        <!-- Filtering -->
        <div class="dropdown-submenu" *ngIf="(column.canFilter || column.canSearch) && status?.filters" [attr.data-qa]="'controls-filter'"
             [ngClass]="{'dropdown-submenu-left': columnIndex > (columnsCount - 5)}" (click)="$event.stopPropagation()">
            <button class="dropdown-item dropdown-submenu-button" [attr.data-qa]="'controls-filter-dropdown'"
                    [ngClass]="{'active' :  status.filters[column.prop]?.hasFilters }">
                <span class="fa-stack fa-stack-groupsort">
                    <i class="fa fa-filter" aria-hidden="true"></i>
                </span> Filters
            </button>
            <div class="dropdown-menu dropdown-submenu-flyout">

                <filters
                    [column]="column"
                    [state]="state"
                    [status]="status"
                    [options]="options"
                    [filterTerms]="filterTerms"
                    (onFiltersUpdated)="modifyFilters($event)">
                </filters>
                
                <div class="pl-2 pt-1 pr-2 pb-0">
                    <button class="btn btn-interact-outline btn-sm mt-2 w-100" (click)="clearFilters(column.prop)" [attr.data-qa]="'controls-filter-clear'">Clear Filters</button>
                </div>
            </div>
        </div>

        <!-- Custom Links -->
        <ng-container *ngIf="column?.controlsCustomLinks?.length">
            <div class="dropdown-divider"></div>
            <!-- Loop through custom links -->
                <ng-container *ngFor="let link of column.controlsCustomLinks">
                    <div class="dropdown-submenu"
                         [ngClass]="{'dropdown-submenu-left': columnIndex > (columnsCount - 5) }"
                         [attr.data-qa]="'controls-custom-link-dropdown'">
                        <button class="dropdown-item" [attr.data-qa]="'controls-custom-link'" [ngClass]="{ 'dropdown-submenu-button' : link?.submenu?.length }">
                            <span class="fa-stack fa-stack-groupsort">
                                <i class="{{link.icon}}" aria-hidden="true" *ngIf="link.icon"></i>
                            </span> {{link.label}}
                        </button>
                        <!-- If submenu links, show flyout -->
                        <div class="dropdown-menu dropdown-submenu-flyout" *ngIf="link?.submenu?.length">
                            <button class="dropdown-item" [attr.data-qa]="'controls-custom-link'" *ngFor="let submenu of link.submenu" (click)="customLinkAction(submenu,column)">
                                <span class="fa-stack fa-stack-groupsort">
                                    <i class="{{submenu.icon}}" aria-hidden="true" *ngIf="submenu.icon"></i>
                                </span> {{submenu.label}}
                            </button>
                        </div>
                    </div>
                </ng-container>
            </ng-container>

            <!-- Deleting -->
            <ng-container *ngIf="column.canDelete">
                <div class="dropdown-divider"></div>
                <button class="dropdown-item" (click)="deleteColumn(columnIndex)" [attr.data-qa]="'controls-delete'">
                    <span class="fa-stack fa-stack-groupsort">
                        <i class="fa fa-remove" aria-hidden="true"></i>
                    </span> Delete This Column
                </button>
            </ng-container>

    </div>
</div>
`,
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
ControlsComponent.ctorParameters = () => [];
ControlsComponent.propDecorators = {
    "state": [{ type: Input },],
    "status": [{ type: Input },],
    "options": [{ type: Input },],
    "column": [{ type: Input },],
    "filterTerms": [{ type: Input },],
    "columnIndex": [{ type: Input },],
    "onStateUpdated": [{ type: Output },],
    "onCustomLinkEvent": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class HeaderRowComponent {
    constructor() {
        // private columnsOriginal;
        this.onColumnsUpdated = new EventEmitter();
        this.onStateUpdated = new EventEmitter();
        this.onCustomLinkEvent = new EventEmitter();
        this.cellContext = {};
        /**
         * During a resize, disable some stuff
         */
        this.reSizing = false;
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * @return {?}
     */
    ngOnChanges() {
        // this.cellContext.column = this.column;
        // this.cellContext.row = this.row;
        // this.cellContext.options = this.options;
        // this.cellContext.value = this.row[this.column.prop];
        // if (this.columns) {
        //		this.columnsOriginal = [...this.columns];
        // }
    }
    /**
     * Pass state changes up from controls component
     * @param {?} event
     * @return {?}
     */
    stateUpdated(event) {
        this.onStateUpdated.emit(event);
    }
    /**
     * Emit a custom link event response up to the parent component
     * @param {?} data
     * @return {?}
     */
    customLinkEvent(data) {
        this.onCustomLinkEvent.emit(data);
    }
    /**
     * Return a unique ID to ngfor to improve performance
     * @param {?} _index
     * @param {?} item - The column
     * @return {?}
     */
    trackColumn(_index, item) {
        return item.$$track;
    }
    /**
     * On a successful drag reorder of the column headers
     * @param {?} type
     * @param {?} columnNewPosition
     * @return {?}
     */
    dragDrop(type, columnNewPosition) {
        // console.log('onReorderSuccess', this.dragStartProp, columnNewPosition);
        // If columns are being dragged before a pinned column, set that column to pinned
        let /** @type {?} */ isPinned = false;
        const /** @type {?} */ payload = { action: 'reorder', type: type, prop: this.dragStartProp, columnIndex: columnNewPosition };
        for (let /** @type {?} */ i = this.columns.length - 1; i >= 0; i--) {
            const /** @type {?} */ column = this.columns[i];
            if (column.pinnedLeft) {
                isPinned = true;
            }
            column.locked = isPinned;
            column.pinnedLeft = isPinned;
        }
        this.onColumnsUpdated.emit(payload);
    }
    /**
     * If the column was resized
     * @param {?} event
     * @param {?} columnIndex
     * @param {?} type
     * @return {?}
     */
    onResizeEnd(event, columnIndex, type) {
        //console.warn('onResizeEnd', columnIndex, type, Math.floor(event.rectangle.width / 2) * 2)
        let /** @type {?} */ width = Math.floor(event.rectangle.width / 2) * 2; // Round to nearest 2 pixels to prevent rendering issues in chrome
        // Min column size 44 px
        if (width < 44) {
            width = 44;
        }
        this.reSizing = false;
        this.onColumnsUpdated.emit({ action: 'resize', columnIndex: columnIndex, type: type, width: width });
    }
}
HeaderRowComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid-header-row',
                template: `<div draggable="false">
    <div class="datagrid-header-cell" [title]="column.label"
         *ngFor="let column of columns; let columnIndex = index; trackBy:trackColumn"
         [ngClass]="{'reorderable': column.canReorder, 'column-last' : columnIndex == columns.length - 1 }"
         [ngStyle]="{'width' : column.$$width + 'px', 'left' : column.$$leftOffset + 'px' }"
         mwlResizable [enableGhostResize]="true" [resizeEdges]="{right: column.canResize || false}" (resizeEnd)="onResizeEnd($event, column.$$index, columnType)"
         [draggable]="column.canReorder" (dragstart)="dragStartProp = column.prop;" (dragover)="$event.preventDefault()" (drop)="dragDrop(columnType, column.$$index)">

        <!-- Column Label -->
        <span class="datagrid-header-cell-label">
            <ng-template #cellTemplate
                         *ngIf="templates && templates[column.prop] && templates[column.prop].templateHeader; else defaults"
                         [ngTemplateOutlet]="templates[column.prop].templateHeader"
                         [ngTemplateOutletContext]="cellContext">
            </ng-template>

            <ng-template #defaults>
                {{column.label}}
            </ng-template>
        </span>

        <!-- Controls Dropdown -->
        <datagrid-controls *ngIf="options.controlsDropdown"
                           [options]="options"
                           [status]="status"
                           [column]="column"
                           [columnIndex]="columnIndex"
                           [filterTerms]="filterTerms"
                           (onStateUpdated)="stateUpdated($event)"
                           (onCustomLinkEvent)="customLinkEvent($event)">
        </datagrid-controls>

        <!-- Resize Handle -->
        <a class="resize-handle" mwlResizeHandle [resizeEdges]="{right: true}" *ngIf="column.canResize" (mousedown)="reSizing = true"></a>

    </div>
</div>`,
                styles: [],
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
HeaderRowComponent.ctorParameters = () => [];
HeaderRowComponent.propDecorators = {
    "columns": [{ type: Input },],
    "state": [{ type: Input },],
    "status": [{ type: Input },],
    "options": [{ type: Input },],
    "scrollProps": [{ type: Input },],
    "filterTerms": [{ type: Input },],
    "columnType": [{ type: Input },],
    "templates": [{ type: Input },],
    "onColumnsUpdated": [{ type: Output },],
    "onStateUpdated": [{ type: Output },],
    "onCustomLinkEvent": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class GroupHeaderComponent {
    constructor() {
        this.onGroupToggled = new EventEmitter();
        this.groupLabel = '';
        this.active = true;
    }
    /**
     * @return {?}
     */
    ngOnChanges() {
        // Everytime new data is passed down, recreate the label
        this.createGroupLabel(this.group, this.options);
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * Toggle group visibility
     * @param {?} group
     * @param {?} $event
     * @return {?}
     */
    toggleGroup(group, $event) {
        $event.preventDefault();
        $event.stopPropagation();
        if (group && group.rows) {
            group.rows.forEach(row => {
                if (!row.$$hidden) {
                    row.$$hidden = true;
                }
                else {
                    row.$$hidden = false;
                }
            });
            this.group.visible = !this.group.visible;
            this.active = !this.active;
            this.onGroupToggled.emit(this.group);
        }
    }
    /**
     * Create the label for the group
     * @param {?} group
     * @param {?} options
     * @return {?}
     */
    createGroupLabel(group, options) {
        // If columnData has been supplied then the app needs to get the group header label from within that data
        if (group &&
            group.columnProp &&
            options.columnData &&
            options.columnData[group.columnProp] &&
            options.columnData[group.columnProp].model) {
            // Sub to the observable
            this.sub = options.columnData[group.columnProp].model.subscribe(model => {
                // If this row property is an array, join all possible settings
                if (options.columnData &&
                    group &&
                    group.label &&
                    Array.isArray(group.label) &&
                    group.columnProp &&
                    group.rows) {
                    const /** @type {?} */ columnData = options.columnData[group.columnProp];
                    const /** @type {?} */ newLabel = [];
                    // Loop through the labels in the group prop
                    group.rows[0][group.columnProp].forEach((id) => {
                        model[columnData.modelSrc].forEach((element) => {
                            if (id === element[columnData.value]) {
                                newLabel.push(element[columnData.label]);
                            }
                        });
                    });
                    this.groupLabel = newLabel.length ? newLabel.join(' & ') : 'Missing Value';
                }
                else {
                    // If this row property is NOT an array
                    this.groupLabel = 'No Value';
                }
            });
        }
        else {
            // No columnData provided
            this.groupLabel = this.group.label ? this.group.label : 'No Value';
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        if (this.sub) {
            this.sub.unsubscribe();
        }
    }
}
GroupHeaderComponent.decorators = [
    { type: Component, args: [{
                selector: 'group-header',
                template: `<div class="datagrid-group-header">
    <div class="datagrid-group-header-cell" [ngStyle]="{'width' : width + 'px' }">
        <a (click)="toggleGroup(group, $event)" [ngClass]="{'active': active }">
            <span class="sticky-left">
                <strong>
                    <span class="red" *ngIf="options?.showGroupColumnName">{{group.columnLabel}}:</span>

                    <ng-container *ngIf="group?.column?.columnType && groupLabel != 'No Value' && groupLabel != 'Missing Value' ">
                        <ng-container [ngSwitch]="group.column.columnType">
                            <!-- Date Time-->
                            <ng-container *ngSwitchCase="'dateTime'">
                                {{groupLabel | date: 'MM/dd/yyyy - h:mm a' }}
                            </ng-container>
                            <!-- Date-->
                            <ng-container *ngSwitchCase="'date'">
                                {{groupLabel | date: 'MM/dd/yyyy' }}
                            </ng-container>
                            <!-- Money -->
                            <ng-container *ngSwitchCase="'money'">
                                {{groupLabel | currency:'USD':true:'1.2-2' }}
                            </ng-container>
                            <!-- Default -->
                            <ng-container *ngSwitchDefault>
                                {{groupLabel}}
                            </ng-container>
                        </ng-container>
                    </ng-container>
                    <ng-container *ngIf="!group?.column?.columnType">
                        {{groupLabel}}
                    </ng-container>
                </strong> <small *ngIf="group?.rows?.length"><em>({{group.rows.length}})</em></small>
            </span>
        </a>
    </div>
</div>

`,
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
GroupHeaderComponent.ctorParameters = () => [];
GroupHeaderComponent.propDecorators = {
    "width": [{ type: Input },],
    "group": [{ type: Input },],
    "options": [{ type: Input },],
    "onGroupToggled": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class RowComponent {
    /**
     * @param {?} elementRef
     */
    constructor(elementRef) {
        this.elementRef = elementRef;
        this.updateDatatable = new EventEmitter();
        this.onRowUpdated = new EventEmitter();
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * @return {?}
     */
    ngOnChanges() { }
    /**
     * Return a unique ID to ngfor to improve performance
     * @param {?} _index
     * @param {?} item - The column
     * @return {?}
     */
    trackColumn(_index, item) {
        return item.$$track;
    }
    /**
     * Perform an action on the main datatable that was requested by lower component
     * @param {?} action
     * @return {?}
     */
    onUpdateDatatable(action) {
        this.updateDatatable.emit(action);
    }
    /**
     * Pass updated field up the component chain
     * @param {?} event
     * @return {?}
     */
    rowUpdated(event) {
        this.onRowUpdated.emit(event);
    }
}
RowComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid-body-row',
                template: `<div class="datagrid-row">
    <ng-container *ngFor="let column of columns; let columnIndex = index; trackBy:trackColumn">

        <datagrid-cell class="datagrid-cell"
                       [ngStyle]="{'left' : column.$$leftOffset + 'px' }"
                       [column]="column"
                       [row]="row"
                       [options]="options"
                       [templates]="templates"
                       (onRowUpdated)="rowUpdated($event)">
        </datagrid-cell>

    </ng-container>
</div>
`,
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
RowComponent.ctorParameters = () => [
    { type: ElementRef, },
];
RowComponent.propDecorators = {
    "columns": [{ type: Input },],
    "options": [{ type: Input },],
    "row": [{ type: Input },],
    "templates": [{ type: Input },],
    "updateDatatable": [{ type: Output },],
    "onRowUpdated": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class CellComponent {
    constructor() {
        this.updateDatatable = new EventEmitter();
        this.onRowUpdated = new EventEmitter();
        /**
         * Use to pass internal data such as columns and rows to the templates projected from the parent component
         */
        this.cellContext = {};
        /**
         * Is the content truncated, IE is the content inside the cell wider than the parent container
         */
        this.truncated = false;
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * @return {?}
     */
    ngOnChanges() {
        if (this.row && this.column.prop) {
            this.checkTruncated();
            this.cellContext.column = this.column;
            this.cellContext.row = this.row;
            this.cellContext.options = this.options;
            this.cellContext.value = this.row[this.column.prop];
        }
    }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        this.checkTruncated();
    }
    /**
     * Check if the content is truncated
     * @return {?}
     */
    checkTruncated() {
        // This is a slightly costly operation since so many cells are present, wrap in setTimeout to release the DOM
        setTimeout(() => {
            if (this.cellData &&
                this.column &&
                this.column.$$width &&
                this.cellData.nativeElement &&
                this.cellData.nativeElement.getBoundingClientRect().width > this.column.$$width) {
                this.truncated = true;
            }
            else {
                this.truncated = false;
            }
        }, 10);
    }
    /**
     * Open the edit note tooltip and set the focus
     * @return {?}
     */
    fieldEdit() {
        if (this.column.canEdit) {
            this.p.open();
            setTimeout(() => this.editBox.nativeElement.focus());
        }
    }
    /**
     * When a field was edited/updated
     * @param {?} event
     * @return {?}
     */
    rowUpdated(event) {
        if (this.column.prop && this.row) {
            const /** @type {?} */ valueOld = this.row[this.column.prop];
            this.row[this.column.prop] = event.target.value;
            const /** @type {?} */ fieldData = {
                valueNew: event.target.value,
                valueOld: valueOld,
                prop: this.column.prop,
                row: this.row,
            };
            // If the data has changed
            if (fieldData.valueNew !== fieldData.valueOld) {
                // Pass data up chain
                this.onRowUpdated.emit(fieldData);
            }
            // Fixes bug with ng-bootstrap not seeing close method
            setTimeout(() => this.p.close());
        }
        event.preventDefault();
        event.stopPropagation();
    }
    /**
     * Perform an action on the main datatable that was requested by lower component
     * @param {?} action
     * @return {?}
     */
    onUpdateDatatable(action) {
        this.updateDatatable.emit(action);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        // Clean up after custom cell templates
        if (this.cellTemplate) {
            this.cellTemplate.clear();
        }
    }
}
CellComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid-cell',
                template: `<div class="datagrid-cell-data"
     title="{{row[column.prop]}}"
     [ngStyle]="{'width' : column.$$width + 'px'}"
     [ngClass]="{'canEdit': column.canEdit }"
     [ngbPopover]="fieldContent"
     #p="ngbPopover"
     container="body"
     placement="top-left"
     triggers="manual"
     (dblclick)="fieldEdit();">

    <div class="datagrid-cell-data-content" #cellData>
        
        <ng-template #cellTemplate
                     *ngIf="templates && templates[column.prop] && templates[column.prop].templateCell; else defaults"
                     [ngTemplateOutlet]="templates[column.prop].templateCell"
                     [ngTemplateOutletContext]="cellContext">
        </ng-template>

        <ng-template #defaults>
            <ng-container [ngSwitch]="column.columnType">

                <!-- Date Time-->
                <ng-container *ngSwitchCase="'dateTime'">
                    {{row[column.prop] | date: 'MM/dd/yyyy - h:mm a' }}
                </ng-container>

                <!-- Date-->
                <ng-container *ngSwitchCase="'date'">
                    {{row[column.prop] | date: 'MM/dd/yyyy' }}
                </ng-container>

                <!-- Money -->
                <ng-container *ngSwitchCase="'money'">
                    {{row[column.prop] | currency:'USD' }}
                </ng-container>

                <!-- Default -->
                <ng-container *ngSwitchDefault>
                    {{row[column.prop]}}
                </ng-container>

            </ng-container>
        </ng-template>

    </div>

  <div class="datagrid-truncated" *ngIf="truncated"></div>
</div>

<ng-template #fieldContent>
  <textarea class="form-control form-control-sm" style="width:400px;" rows="3" #editBox autofocus
            (dblclick)="$event.target.select();"
            (blur)="rowUpdated($event);"
            (keyup.enter)="rowUpdated($event);">{{row[column.prop]}}</textarea>
</ng-template>
`,
                providers: [],
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
CellComponent.ctorParameters = () => [];
CellComponent.propDecorators = {
    "column": [{ type: Input },],
    "row": [{ type: Input },],
    "options": [{ type: Input },],
    "templates": [{ type: Input },],
    "updateDatatable": [{ type: Output },],
    "onRowUpdated": [{ type: Output },],
    "p": [{ type: ViewChild, args: ['p',] },],
    "editBox": [{ type: ViewChild, args: ['editBox',] },],
    "cellData": [{ type: ViewChild, args: ['cellData',] },],
    "cellTemplate": [{ type: ViewChild, args: ['cellTemplate', { read: ViewContainerRef },] },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class InfoComponent {
    constructor() {
        this.onReset = new EventEmitter();
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * Reset one of the datatable controls
     * @param {?} resetType
     * @return {?}
     */
    reset(resetType) {
        this.onReset.emit(resetType);
    }
}
InfoComponent.decorators = [
    { type: Component, args: [{
                selector: 'datagrid-info',
                template: `<div class="datagrid-info-row" *ngIf="columnsMapped && state">
    <span *ngIf="state.groups.length && columnsMapped[state?.groups[0].prop]">
        Grouped by <strong>{{columnsMapped[state?.groups[0].prop].label}}, {{state.groups[0].dir | uppercase}}</strong>
        <a (click)="reset('groups')" *ngIf="options.controlsDropdown" title="Remove Grouping"><i class="fa fa-remove"></i></a>
    </span>
    <ng-container *ngIf="state.groups.length && state.sorts.length"> | </ng-container>
    <span *ngIf="state.sorts.length && columnsMapped[state?.sorts[0].prop]">
        Sorted by <strong>{{columnsMapped[state?.sorts[0].prop].label}}, {{state.sorts[0].dir | uppercase}}</strong>
        <a (click)="reset('sorts')" *ngIf="options.controlsDropdown" title="Remove Sorting"><i class="fa fa-remove"></i></a>
    </span>
    <ng-container *ngIf="state.sorts.length && state.filters.length"> | </ng-container>
    <span *ngIf="state.filters.length">
        <strong>{{state?.filters?.length}} Filter/s Active</strong> {{state?.info?.rowsVisible}}/{{state?.info?.rowsTotal}} Rows Visible:
        <small *ngFor="let filter of state.filters">
            ({{columnsMapped[filter.prop].label}}
            <ng-container *ngIf="filter.operator == 'eq'">=</ng-container>
            <ng-container *ngIf="filter.operator == 'contains'">contains</ng-container>
            {{filter.value}})
        </small>
        <a (click)="reset('filters')" *ngIf="options.controlsDropdown" title="Remove All Filters"><i class="fa fa-remove"></i></a>
    </span>
    <span *ngIf="options?.filterGlobal?.term"> | Global Filter: <strong>{{options.filterGlobal.term}}</strong></span>
</div>
`,
                changeDetection: ChangeDetectionStrategy.OnPush,
            },] },
];
/** @nocollapse */
InfoComponent.ctorParameters = () => [];
InfoComponent.propDecorators = {
    "state": [{ type: Input },],
    "options": [{ type: Input },],
    "columnsMapped": [{ type: Input },],
    "onReset": [{ type: Output },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class DatagridModule {
    /**
     * @return {?}
     */
    static forRoot() {
        return {
            ngModule: DatagridModule,
            providers: [DataGridService],
        };
    }
}
DatagridModule.decorators = [
    { type: NgModule, args: [{
                imports: [CommonModule, NgbModule.forRoot(), FormsModule, ResizableModule],
                declarations: [
                    DataGridComponent,
                    RowComponent,
                    GroupHeaderComponent,
                    HeaderComponent,
                    ControlsComponent,
                    FiltersComponent,
                    BodyComponent,
                    CellComponent,
                    InfoComponent,
                    HeaderRowComponent,
                    DataTableColumnDirective,
                    DataGridColumnCellDirective,
                    DataGridColumnHeaderDirective,
                ],
                providers: [DataGridService],
                exports: [DataGridComponent, DataTableColumnDirective, DataGridColumnCellDirective, DataGridColumnHeaderDirective],
            },] },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * Generated bundle index. Do not edit.
 */

export { DatagridModule, BodyComponent as ɵk, CellComponent as ɵl, GroupHeaderComponent as ɵg, RowComponent as ɵf, DataGridComponent as ɵa, ControlsComponent as ɵi, FiltersComponent as ɵj, HeaderComponent as ɵh, HeaderRowComponent as ɵn, InfoComponent as ɵm, DataGridColumnCellDirective as ɵc, DataGridColumnHeaderDirective as ɵd, DataTableColumnDirective as ɵb, DataGridService as ɵe };
//# sourceMappingURL=mello-labs-datagrid.js.map
